from __future__ import annotations

import copy
import os
import tempfile
import unittest

from mock import AsyncMock, MagicMock

from iambic.core.context import ctx
from iambic.core.models import ProposedChangeType, TemplateChangeDetails
from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
    GoogleProject,
    GoogleWorkspaceConfig,
)


class TestGroupTemplateApply(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        from iambic.plugins.v0_1_0.google_workspace.group.models import (
            GoogleWorkspaceGroupTemplate,
            GroupMember,
            GroupProperties,
        )

        workspaces = []
        self.google_workspace_config = GoogleWorkspaceConfig(workspaces=workspaces)
        self.temp_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.temp_dir.cleanup)
        self.temp_file_path = os.path.join(self.temp_dir.name, "test_path.yaml")
        self.template = GoogleWorkspaceGroupTemplate(
            template_type="google:group",
            file_path=self.temp_file_path,
            properties=GroupProperties(
                name="test_group",
                domain="example.com",
                email="test@example.com",
                description="test_description",
                members=[GroupMember(email="testuser@example.com")],
            ),
        )
        self.template_2 = GoogleWorkspaceGroupTemplate(
            template_type="google:group",
            file_path=self.temp_file_path,
            properties=GroupProperties(
                name="test_group",
                domain="example2.com",
                email="test@example2.com",
                description="test_description",
                members=[GroupMember(email="testuser@example2.com")],
            ),
        )

        self.google_project1 = GoogleProject(
            project_id="google_project1",
            subjects=[
                {"domain": "example.com", "service_account": "google_project1_sa"}
            ],
            type="google",
            private_key_id="google_project1_private_key_id",
            private_key="google_project1_private_key",
            client_email="google_project1_client_email",
            client_id="google_project1_client_id",
            auth_uri="google_project1_auth_uri",
            token_uri="google_project1_token_uri",
            auth_provider_x509_cert_url="google_project1_auth_provider_x509_cert_url",
            client_x509_cert_url="google_project1_client_x509_cert_url",
        )
        self.google_project2 = GoogleProject(
            project_id="google_project2",
            subjects=[
                {"domain": "example2.com", "service_account": "google_project2_sa"}
            ],
            type="google",
            private_key_id="google_project2_private_key_id",
            private_key="google_project2_private_key",
            client_email="google_project2_client_email",
            client_id="google_project2_client_id",
            auth_uri="google_project2_auth_uri",
            token_uri="google_project2_token_uri",
            auth_provider_x509_cert_url="google_project2_auth_provider_x509_cert_url",
            client_x509_cert_url="google_project2_client_x509_cert_url",
        )
        mock_req = MagicMock()
        mock_req.execute = MagicMock(
            return_value={
                "name": "test_group",
                "email": "test@example.com",
                "description": "test_description",
            }
        )
        mock_service = MagicMock()
        mock_service.groups.return_value.get.return_value = mock_req
        object.__setattr__(
            self.google_project1,
            "get_service_connection",
            AsyncMock(return_value=mock_service),
        )
        object.__setattr__(
            self.google_project2,
            "get_service_connection",
            AsyncMock(return_value=mock_service),
        )
        self.google_workspace_config.workspaces = [
            self.google_project1,
            self.google_project2,
        ]

    def tearDown(self):
        ctx.eval_only = False

    async def test_apply_with_projects(self):
        ctx.eval_only = True
        google_workspace_config = copy.deepcopy(self.google_workspace_config)
        template_changes = await self.template.apply(google_workspace_config)

        self.assertIsInstance(template_changes, TemplateChangeDetails)
        self.assertEqual(template_changes.resource_id, "test@example.com")
        self.assertEqual(template_changes.resource_type, "google:group")
        self.assertEqual(template_changes.template_path, self.temp_file_path)
        self.assertEqual(len(template_changes.proposed_changes), 1)
        template_changes = await self.template.apply(google_workspace_config)
        self.assertEqual(
            google_workspace_config.workspaces[
                0
            ].get_service_connection.return_value.groups.return_value.get.call_count,
            2,
        )

    async def test_apply_to_account_new_group(self):
        ctx.eval_only = True
        # Set up mock for get_group function to return None (group doesn't exist)
        with unittest.mock.patch(
            "iambic.plugins.v0_1_0.google_workspace.group.models.get_group",
            new_callable=AsyncMock,
        ) as mock_get_group:
            mock_get_group.side_effect = [None, None, self.template]

            # Set up mock for create_group function to simulate group creation
            with unittest.mock.patch(
                "iambic.plugins.v0_1_0.google_workspace.group.models.create_group",
                new_callable=AsyncMock,
            ) as mock_create_group:
                # Call the _apply_to_account method
                change_details = await self.template._apply_to_account(
                    self.google_project1
                )

                # Test that the create_group function not called
                mock_create_group.assert_not_called()

                # Test that the correct changes were proposed
                # Replace with specific assertions based on your implementation
                self.assertEqual(len(change_details.proposed_changes), 1)
                proposed_change = change_details.proposed_changes[0]
                self.assertEqual(proposed_change.change_type, ProposedChangeType.CREATE)
                self.assertEqual(
                    proposed_change.resource_id, self.template.properties.email
                )
                self.assertEqual(
                    proposed_change.resource_type,
                    self.template.properties.resource_type,
                )

                ctx.eval_only = False

                change_details = await self.template._apply_to_account(
                    self.google_project1
                )

                # Test that the create_group function was called
                mock_create_group.assert_called_once_with(
                    id=self.template.properties.email,
                    domain=self.template.properties.domain,
                    email=self.template.properties.email,
                    name=self.template.properties.name,
                    description=self.template.properties.description,
                    google_project=self.google_project1,
                )

                self.assertEqual(len(change_details.proposed_changes), 1)
                proposed_change = change_details.proposed_changes[0]
                self.assertEqual(proposed_change.change_type, ProposedChangeType.CREATE)
                self.assertEqual(
                    proposed_change.resource_id, self.template.properties.email
                )
                self.assertEqual(
                    proposed_change.resource_type,
                    self.template.properties.resource_type,
                )



from __future__ import annotations

import asyncio
import json
from itertools import chain
from typing import TYPE_CHECKING, Union

from deepdiff import DeepDiff

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import aio_wrapper, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, paginated_search

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.models import AWSAccount


async def get_group_inline_policy_names(group_name: str, iam_client):
    return await paginated_search(
        iam_client.list_group_policies, "PolicyNames", GroupName=group_name
    )


async def list_groups(iam_client):
    return await paginated_search(iam_client.list_groups, "Groups")


async def list_users_in_group(group_name: str, iam_client):
    return await paginated_search(iam_client.get_group, "Users", GroupName=group_name)


async def get_group_policy(group_name: str, policy_name: str, iam_client):
    return await boto_crud_call(
        iam_client.get_group_policy, GroupName=group_name, PolicyName=policy_name
    )


async def get_group_inline_policies(
    group_name: str, iam_client, as_dict: bool = True
) -> Union[list, dict]:
    policy_names = await get_group_inline_policy_names(group_name, iam_client)
    policies = await asyncio.gather(
        *[
            get_group_policy(group_name, policy_name, iam_client)
            for policy_name in policy_names
        ]
    )
    if as_dict:
        return {policy["PolicyName"]: policy["PolicyDocument"] for policy in policies}
    else:
        return [
            {"PolicyName": policy["PolicyName"], **policy["PolicyDocument"]}
            for policy in policies
        ]


async def get_group_managed_policies(
    group_name: str, iam_client
) -> list[dict[str, str]]:
    marker: dict[str, str] = {}
    policies = []

    while True:
        response = await boto_crud_call(
            iam_client.list_attached_group_policies, GroupName=group_name, **marker
        )
        policies.extend(response["AttachedPolicies"])

        if response["IsTruncated"]:
            marker["Marker"] = response["Marker"]
        else:
            break

    return policies


async def get_group(group_name: str, iam_client, include_policies: bool = True) -> dict:
    try:
        current_group = (
            await boto_crud_call(iam_client.get_group, GroupName=group_name)
        )["Group"]
        if include_policies:
            current_group["ManagedPolicies"] = await get_group_managed_policies(
                group_name, iam_client
            )
            current_group["InlinePolicies"] = await get_group_inline_policies(
                group_name, iam_client, as_dict=False
            )
    except iam_client.exceptions.NoSuchEntityException:
        current_group = {}

    return current_group


async def get_group_across_accounts(
    aws_accounts: list[AWSAccount], group_name: str, include_policies: bool = True
) -> dict:
    async def get_group_for_account(aws_account: AWSAccount):
        iam_client = await aws_account.get_boto3_client("iam")
        return {
            aws_account.account_id: await get_group(
                group_name, iam_client, include_policies
            )
        }

    account_on_groups = await asyncio.gather(
        *[get_group_for_account(aws_account) for aws_account in aws_accounts]
    )
    return {
        account_id: group
        for resp in account_on_groups
        for account_id, group in resp.items()
    }


async def apply_group_managed_policies(
    group_name,
    iam_client,
    template_policies: list[dict],
    existing_policies: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []
    template_policies = [policy["PolicyArn"] for policy in template_policies]
    existing_managed_policies = [policy["PolicyArn"] for policy in existing_policies]

    # Create new managed policies
    new_managed_policies = [
        policy_arn
        for policy_arn in template_policies
        if policy_arn not in existing_managed_policies
    ]
    if new_managed_policies:
        log_str = "New managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Attaching managed policies..."

        for policy_arn in new_managed_policies:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.ATTACH,
                    resource_type="aws:policy_document",
                    resource_id=policy_arn,
                    attribute="managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    iam_client.attach_group_policy,
                    GroupName=group_name,
                    PolicyArn=policy_arn,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, managed_policies=new_managed_policies, **log_params)

    # Delete existing managed policies not in template
    existing_managed_policies = [
        policy_arn
        for policy_arn in existing_managed_policies
        if policy_arn not in template_policies
    ]
    if existing_managed_policies:
        log_str = "Stale managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Detaching managed policies..."

        for policy_arn in existing_managed_policies:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DETACH,
                    resource_type="aws:policy_document",
                    resource_id=policy_arn,
                    attribute="managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    iam_client.detach_group_policy,
                    GroupName=group_name,
                    PolicyArn=policy_arn,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, managed_policies=existing_managed_policies, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_group_inline_policies(
    group_name,
    iam_client,
    template_policies: list[dict],
    existing_policies: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    apply_tasks = []
    delete_tasks = []
    plan_response = []
    template_policy_map = {
        policy["PolicyName"]: {k: v for k, v in policy.items() if k != "PolicyName"}
        for policy in template_policies
    }
    existing_policy_map = {
        policy["PolicyName"]: {k: v for k, v in policy.items() if k != "PolicyName"}
        for policy in existing_policies
    }

    for policy_name in existing_policy_map.keys():
        if not template_policy_map.get(policy_name):
            log_str = "Stale inline policies discovered."
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DELETE,
                    resource_type="aws:policy_document",
                    resource_id=policy_name,
                    attribute="inline_policies",
                )
            ]
            plan_response.extend(proposed_changes)

            if ctx.execute:
                log_str = f"{log_str} Removing inline policy..."

                apply_awaitable = boto_crud_call(
                    iam_client.delete_group_policy,
                    GroupName=group_name,
                    PolicyName=policy_name,
                )
                delete_tasks.append(
                    plugin_apply_wrapper(apply_awaitable, proposed_changes)
                )
            log.debug(log_str, policy_name=policy_name, **log_params)

    for policy_name, policy_document in template_policy_map.items():
        existing_policy_doc = existing_policy_map.get(policy_name)
        policy_drift = None
        if existing_policy_doc:
            policy_drift = await aio_wrapper(
                DeepDiff,
                existing_policy_doc,
                policy_document,
                ignore_order=True,
                report_repetition=True,
            )

            # DeepDiff will return type changes as actual type functions and not strings,
            # and this will cause json serialization to fail later on when we process
            # the proposed changes. We force type changes to strings here.
            policy_drift = json.loads(policy_drift.to_json())

        if not existing_policy_doc or policy_drift:
            if policy_drift:
                log_params["policy_drift"] = policy_drift
                boto_action = "Updating"
                resource_existence = "Stale"
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.UPDATE,
                        resource_type="aws:policy_document",
                        resource_id=policy_name,
                        attribute="inline_policies",
                        change_summary=policy_drift,
                        current_value=existing_policy_doc,
                        new_value=policy_document,
                    )
                ]
            else:
                boto_action = "Creating"
                resource_existence = "New"
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_type="aws:policy_document",
                        resource_id=policy_name,
                        attribute="inline_policies",
                        new_value=policy_document,
                    )
                ]
            plan_response.extend(proposed_changes)

            log_str = f"{resource_existence} inline policies discovered."
            if ctx.execute and policy_document:
                log_str = f"{log_str} {boto_action} inline policy..."
                apply_awaitable = boto_crud_call(
                    iam_client.put_group_policy,
                    GroupName=group_name,
                    PolicyName=policy_name,
                    PolicyDocument=json.dumps(policy_document),
                )
                apply_tasks.append(
                    plugin_apply_wrapper(apply_awaitable, proposed_changes)
                )

            log.debug(log_str, policy_name=policy_name, **log_params)

    if apply_tasks or delete_tasks:
        results: list[list[ProposedChange]] = []
        if delete_tasks:
            results.extend(await asyncio.gather(*delete_tasks))
            if apply_tasks:
                # Wait for the policy deletion to propagate before applying new policies
                # Otherwise a max policy limit error may be thrown
                await asyncio.sleep(3)

        if apply_tasks:
            results.extend(await asyncio.gather(*apply_tasks))

        return list(chain.from_iterable(results))
    else:
        return plan_response


async def delete_iam_group(group_name: str, iam_client, log_params: dict):
    tasks = []
    # Remove users from group
    attached_users = await list_users_in_group(group_name, iam_client)
    if attached_users:
        attached_users = [attached_user["UserName"] for attached_user in attached_users]
        log.debug(
            "Removing users from group.", attached_users=attached_users, **log_params
        )
        for attached_user in attached_users:
            tasks.append(
                boto_crud_call(
                    iam_client.remove_user_from_group,
                    GroupName=group_name,
                    UserName=attached_user,
                )
            )
        await asyncio.gather(*tasks)

    tasks = []
    # Detach managed policies
    managed_policies = await get_group_managed_policies(group_name, iam_client)
    if managed_policies:
        managed_policies = [policy["PolicyArn"] for policy in managed_policies]
        log.debug(
            "Detaching managed policies.",
            managed_policies=managed_policies,
            **log_params,
        )
        for policy in managed_policies:
            tasks.append(
                boto_crud_call(
                    iam_client.detach_group_policy,
                    GroupName=group_name,
                    PolicyArn=policy,
                )
            )

    # Delete inline policies
    inline_policies = await get_group_inline_policies(group_name, iam_client)
    if inline_policies:
        inline_policies = list(inline_policies.keys())
        log.debug(
            "Deleting inline policies.", managed_policies=inline_policies, **log_params
        )
        for policy_name in inline_policies:
            tasks.append(
                boto_crud_call(
                    iam_client.delete_group_policy,
                    GroupName=group_name,
                    PolicyName=policy_name,
                )
            )

    if managed_policies or inline_policies:
        # Actually perform the deletion of Managed & Inline policies
        await asyncio.gather(*tasks)

    # Now that everything has been removed from the group, delete the group itself
    await boto_crud_call(iam_client.delete_group, GroupName=group_name)



from __future__ import annotations

import json
import pathlib
from typing import Any, Dict, List, Optional, Set

from dictdiffer import diff
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    ProposedChange,
    ProposedChangeType,
    TemplateChangeDetails,
)
from iambic.core.utils import camel_to_snake


class ProposedChangeDiff(ProposedChange):
    diff: Optional[str]
    diff_resolved: Optional[str]

    field_map = {
        "inline_policies": "InlinePolicies",
        "managed_policies": "ManagedPolicies",
        "policy_document": "PolicyDocument",
        "tags": "Tags",
        "assume_role_policy_document": "AssumeRolePolicyDocument",
        "permission_boundary": "PermissionBoundary",
        "customer_managed_policies": "CustomerManagedPolicies",
        "group_name": "GroupName",
        "group_email": "GroupEmail",
        "assignments": "Assignment",
        "domain": "Domain",
        "group": "Group",
        "groups": "Group",
        "users": "User",
        "user": "User",
    }

    def __init__(self, proposed_change: ProposedChange) -> None:
        super().__init__(**proposed_change.dict())
        try:
            object_attribute = camel_to_snake(self.attribute)
        except TypeError:
            object_attribute = None
        if self.current_value is None:
            self.current_value = {}
        if self.new_value is None:
            self.new_value = {}
        if isinstance(self.current_value, dict):
            self.diff = list(
                diff(self.current_value.get(object_attribute, {}), self.new_value)
            )
        else:
            self.diff = list(diff(self.current_value, self.new_value))

    @property
    def diff_plus_minus(self) -> List[str]:
        diff_plus_minus = ""
        for x in self.diff:
            label = self.attribute
            if x[0] == "change":
                if isinstance(x[2], list) or isinstance(x[2], tuple):
                    change_from = x[2][0]
                    change_to = x[2][1]
                else:
                    change_from = x[2]
                    change_to = x[2]
                if change_to:
                    diff_plus_minus += f"{label}:\n-(From)\n{json.dumps(change_from, indent=2)}\n+(To)\n{json.dumps(change_to, indent=2)}"
                else:
                    diff_plus_minus += (
                        f"{label}:\n-(Remove)\n{json.dumps(change_from, indent=2)}"
                    )
                diff_plus_minus.rstrip("\n")
            elif x[0] == "add":
                diff_plus_minus += (
                    f"{label}:\n+(Add)\n{json.dumps([y[1] for y in x[2]], indent=2)}"
                )
                diff_plus_minus.rstrip("\n")
            elif x[0] == "remove":
                diff_plus_minus += (
                    f"{label}:\n-(Remove)\n{json.dumps([y[1] for y in x[2]], indent=2)}"
                )
                diff_plus_minus.rstrip("\n")
        return diff_plus_minus


class ApplicableChange(PydanticBaseModel):
    account: Optional[str]
    change: Optional[ProposedChange]
    template_change: Optional[TemplateChangeDetails]
    template_name: Optional[str]
    resource_id: Optional[str]
    resource_type: Optional[str]

    def __hash__(self):
        return hash((self.resource_id, self.resource_type))

    def __init__(
        self,
        change: ProposedChange,
        template_change: TemplateChangeDetails,
        **data: Any,
    ) -> None:
        super().__init__(
            change=ProposedChangeDiff(change), template_change=template_change, **data
        )
        self.template_name = pathlib.Path(template_change.template_path).name


class AccountSummary(PydanticBaseModel):
    account: Optional[str] = Field(default="NONE")
    count: Optional[int]
    num_changes: Optional[int]
    changes: Optional[List[ProposedChange]]

    def __hash__(self):
        return hash((self.account, self.num_changes))

    @classmethod
    def compile(
        cls, account: str, count: int, changes: List[ProposedChange], **data: Any
    ) -> None:
        instance = cls()
        instance.account = account
        instance.count = count
        instance.num_changes = len(changes)
        instance.changes = changes
        return instance


class TemplateSummary(PydanticBaseModel):
    template_path: Optional[str]
    template_name: Optional[str]
    count: Optional[int]
    num_accounts: Optional[int]
    accounts: Optional[List[AccountSummary]]

    def __hash__(self):
        return hash(self.template_path)

    @classmethod
    def compile(
        cls,
        template_path: str,
        template_name: str,
        count: int,
        changes: List[ProposedChange],
        **data: Any,
    ) -> None:
        instance = cls()
        instance.template_path = template_path
        instance.template_name = template_name
        instance.count = count
        instance.num_accounts = len(set([x.account for x in changes]))
        accounts = set(x.account for x in changes)
        instance.accounts = [
            AccountSummary.compile(
                account=account,
                count=len(changes),
                changes=[x for x in changes if x.account == account],
            )
            for account in accounts
        ]
        return instance


def get_applicable_changes(
    template_changes: List[TemplateChangeDetails],
    proposed_change_type: str,
    attribute: str = "proposed_changes",
) -> Any:
    """Compile applicable changes as a list of ApplicableChange objects.

    :param template_changes: list of TemplateChangeDetails objects
    :param proposed_change_type: one of ProposedChangeType values
    :param attribute: str. is either "proposed_changes" or "exceptions_seen"
    :return: set of ApplicableChange
    """

    def _get_annotated_change(
        change: ProposedChange,
        template_change: TemplateChangeDetails,
        account: str = "NONE",
    ) -> ApplicableChange:
        return ApplicableChange(
            account=account,
            change=change,
            template_change=template_change,
            resource_id=change.resource_id,
            resource_type=change.resource_type,
        )

    applicable_changes: Set[ApplicableChange] = set()
    for template_change in template_changes:
        for proposed_change in getattr(template_change, attribute, []):
            if isinstance(proposed_change, AccountChangeDetails):
                # If proposed change is a list of AccountChangeDetails, we need to iterate through those
                for account_change in proposed_change.proposed_changes:
                    if account_change.change_type.value == proposed_change_type:
                        applicable_changes.add(
                            _get_annotated_change(
                                account_change, template_change, proposed_change.account
                            )
                        )
            else:
                if proposed_change.change_type.value == proposed_change_type:
                    # If proposed change is a single change, we can just append it
                    applicable_changes.add(
                        _get_annotated_change(proposed_change, template_change)
                    )

    return set(applicable_changes)  # collapse across accounts and no accounts


class ActionSummary(PydanticBaseModel):
    action: Optional[str]
    count: Optional[int]
    num_templates: Optional[int]
    templates: Optional[List[TemplateSummary]]

    @classmethod
    def compile_proposed_changes(
        cls, template_changes: List[TemplateChangeDetails], proposed_change_type: str
    ) -> Any:
        """Compile a list of TemplateChangeDetails into a list of TemplateSummary objects.

        :param resources_changes: list of TemplateChangeDetails objects
        :returns: None
        """
        applicable_changes = get_applicable_changes(
            template_changes, proposed_change_type, attribute="proposed_changes"
        )
        log.debug(f"Found {len(applicable_changes)} applicable changes")

        instance = cls(
            action=proposed_change_type, count=len(applicable_changes), templates=[]
        )
        templates = set(
            [
                TemplateSummary.compile(
                    template_path=x.template_change.template_path,
                    template_name=x.template_name,
                    count=1,
                    changes=[
                        y
                        for y in applicable_changes
                        if y.template_change.template_path
                        == x.template_change.template_path
                    ],
                )
                for x in applicable_changes
            ]
        )
        instance.templates = templates
        instance.num_templates = len(templates)

        return instance


class ExceptionSummary(PydanticBaseModel):
    action: Optional[str]
    count: Optional[int]
    num_templates: Optional[int]
    templates: Optional[List[TemplateSummary]]

    @classmethod
    def compile_exceptions_seen(
        cls, template_changes: List[TemplateChangeDetails], proposed_change_type: str
    ) -> Any:
        exceptions = get_applicable_changes(
            template_changes, proposed_change_type, attribute="exceptions_seen"
        )
        log.debug(f"Found {len(exceptions)} exceptions")

        instance = cls(
            action=proposed_change_type,
            count=len(exceptions),
            num_templates=0,
            templates=[],
        )
        templates = set(
            [
                TemplateSummary.compile(
                    template_path=x.template_change.template_path,
                    template_name=x.template_name,
                    count=1,
                    changes=[
                        y
                        for y in exceptions
                        if y.template_change.template_path
                        == x.template_change.template_path
                    ],
                )
                for x in exceptions
            ]
        )
        instance.templates = templates
        instance.num_templates = len(templates)

        return instance


class ActionSummaries(PydanticBaseModel):
    num_actions: Optional[int]
    num_templates: Optional[int]
    num_accounts: Optional[int]
    num_exceptions: Optional[int]
    action_summaries: Optional[List[ActionSummary]]
    exceptions: Optional[List[ExceptionSummary]]

    @classmethod
    def compile(cls, changes: List[TemplateChangeDetails]):
        instance = cls()
        instance.action_summaries = [
            ActionSummary.compile_proposed_changes(changes, x)
            for x in list([e.value for e in ProposedChangeType])
        ]
        instance.num_actions = sum(
            [1 for x in instance.action_summaries if x.count > 0]
        )
        instance.num_templates = sum(
            [len(x.templates) for x in instance.action_summaries]
        )
        accounts = set(
            [
                g.account
                for y in instance.action_summaries
                for z in y.templates
                for g in z.accounts
            ]
        )
        instance.num_accounts = len(accounts)
        instance.exceptions = [
            ExceptionSummary.compile_exceptions_seen(changes, x)
            for x in list([e.value for e in ProposedChangeType])
        ]
        instance.num_exceptions = sum([1 for x in instance.exceptions if x.count > 0])
        return instance


def get_template_data(resources_changes: List[TemplateChangeDetails]) -> Dict[str, Any]:
    """Convert TemplateChangeDetails into a format that is oriented in this format.

    * Action (Add, Delete, Modify)
    * Template Name (Count of <action>)
    * Account Name (Count of <action>)
    * Changes for Account (Count of <action>)
    * Table of proposed changes

    For each, the corresponding jinja2 templates are:

    * Action: templates/actions.jinja2
    * Template Name: templates/template.jinja2
    * Account Name: templates/accounts.jinja2
    * Changes for Account: templates/resource_change.jinja2

    There is also a jinja template to display all exceptions:

    * Exceptions: templates/exception_details.jinja2

    :param resources_changes: list of TemplateChangeDetails objects
    :returns: Dict[str, Any]
    """
    return ActionSummaries.compile(resources_changes)



from __future__ import annotations

import asyncio
import itertools
import os
from collections import defaultdict
from typing import TYPE_CHECKING, Union

import aiofiles

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, normalize_dict_keys, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import PermissionSetMessageDetails
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.models import (
    AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE,
    AwsIdentityCenterPermissionSetTemplate,
    PermissionSetProperties,
)
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.utils import (
    enrich_permission_set_details,
    get_permission_set_details,
    get_permission_set_users_and_groups,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.template_generation import (
    base_group_str_attribute,
    group_dict_attribute,
    group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.utils import (
    calculate_import_preference,
    get_aws_account_map,
)

# TODO: Update all grouping functions to support org grouping once multiple orgs with IdentityCenter is functional
# TODO: Update partial import to support permission set only being deleted on a single org

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["identity_center", "permission_set"]


def get_template_dir(base_dir: str) -> str:
    repo_dir = os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR)
    os.makedirs(repo_dir, exist_ok=True)
    return repo_dir


def get_templated_permission_set_file_path(
    permission_set_dir: str,
    permission_set_name: str,
):
    file_name = (
        permission_set_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )
    return str(os.path.join(permission_set_dir, f"{file_name}.yaml"))


async def gather_permission_set_names(
    aws_accounts: list[AWSAccount],
    detect_messages: list[PermissionSetMessageDetails],
) -> list[str]:
    sso_admin_client_map = {
        aws_account.account_id: (
            await aws_account.get_boto3_client(
                "sso-admin",
                region_name=aws_account.identity_center_details.region.value,
            )
        )
        for aws_account in aws_accounts
    }

    permission_sets = await asyncio.gather(
        *[
            get_permission_set_details(
                sso_admin_client_map[message.account_id],
                message.instance_arn,
                message.permission_set_arn,
            )
            for message in detect_messages
        ]
    )

    return list(
        set(
            [
                permission_set["Name"]
                for permission_set in permission_sets
                if permission_set
            ]
        )
    )


async def generate_permission_set_resource_file(
    identity_center_client,
    account_id: str,
    instance_arn: str,
    permission_set: dict,
    user_map: dict,
    group_map: dict,
    account_resource_dir: str,
) -> dict:
    permission_set = await enrich_permission_set_details(
        identity_center_client, instance_arn, permission_set
    )
    permission_set["assignments"] = await get_permission_set_users_and_groups(
        identity_center_client,
        instance_arn,
        permission_set["PermissionSetArn"],
        user_map,
        group_map,
    )
    file_path = os.path.join(account_resource_dir, f'{permission_set["Name"]}.json')
    response = dict(
        account_id=account_id, permission_set=permission_set, file_path=file_path
    )
    await resource_file_upsert(file_path, permission_set)

    return response


def _sorted_and_clean_access_rules(unsorted_rules: list) -> list:
    # access_rules has no semantic ordering, we sort it explicitly to ensure the template
    # generation is deterministic.
    sorted_access_rules = sorted(
        unsorted_rules, key=lambda rule: rule["account_rule_key"]
    )
    for rule in sorted_access_rules:
        del rule["account_rule_key"]
    return sorted_access_rules


async def create_templated_permission_set(  # noqa: C901
    aws_account_map: dict[str, AWSAccount],
    permission_set_name: str,
    permission_set_refs: list[dict],
    permission_set_dir: str,
    existing_template_map: dict,
) -> Union[AwsIdentityCenterPermissionSetTemplate, None]:
    account_id_to_permissionn_set_map = {}
    num_of_accounts = len(permission_set_refs)
    for permission_set_ref in permission_set_refs:
        async with aiofiles.open(permission_set_ref["file_path"], mode="r") as f:
            content_dict = json.loads(await f.read())
            account_id_to_permissionn_set_map[
                permission_set_ref["account_id"]
            ] = normalize_dict_keys(content_dict)

    # calculate preference based on existing template
    prefer_templatized = calculate_import_preference(
        existing_template_map.get(permission_set_name)
    )

    # Generate the params used for attribute creation
    template_params = {"identifier": permission_set_name, "access_rules": []}
    template_properties = {"name": permission_set_name}
    description_resources = list()
    relay_state_resources = list()
    customer_managed_policy_ref_resources = list()
    managed_policy_resources = list()
    inline_policy_resources = list()
    permissions_boundary_resources = list()
    tag_resources = list()
    session_duration_resources = list()

    for account_id, permission_set_dict in account_id_to_permissionn_set_map.items():
        if session_duration := permission_set_dict.get("session_duration"):
            session_duration_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": session_duration}],
                }
            )

        if managed_policies := permission_set_dict.get("attached_managed_policies"):
            managed_policy_resources.append(
                {
                    "account_id": account_id,
                    "resources": [
                        {"resource_val": {"arn": mp["arn"]}} for mp in managed_policies
                    ],
                }
            )

        if customer_managed_policy_refs := permission_set_dict.get(
            "customer_managed_policy_references"
        ):
            customer_managed_policy_ref_resources.append(
                {
                    "account_id": account_id,
                    "resources": [
                        {"resource_val": customer_managed_policy_ref}
                        for customer_managed_policy_ref in customer_managed_policy_refs
                    ],
                }
            )

        if permissions_boundary := permission_set_dict.get("permissions_boundary"):
            permissions_boundary_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": permissions_boundary}],
                }
            )

        if tags := permission_set_dict.get("tags"):
            tag_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": tag} for tag in tags],
                }
            )

        if description := permission_set_dict.get("description"):
            description_resources.append(
                {"account_id": account_id, "resources": [{"resource_val": description}]}
            )

        if inline_policy := permission_set_dict.get("inline_policy"):
            inline_policy_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": inline_policy}],
                }
            )

        if relay_state := permission_set_dict.get("relay_state"):
            relay_state_resources.append(
                {"account_id": account_id, "resources": [{"resource_val": relay_state}]}
            )

        aws_account = aws_account_map[account_id]
        account_rules = {}
        for assignment_type in ["user", "group"]:
            for assignment_id, details in (
                permission_set_dict.get("assignments", {})
                .get(assignment_type, {})
                .items()
            ):
                accounts = sorted(details["accounts"])
                account_rule_key = "_".join(accounts)
                if not account_rules.get(account_rule_key):
                    if all(
                        account in accounts
                        for account in aws_account.identity_center_details.org_account_map.keys()
                    ):
                        accounts = ["*"]
                    else:
                        accounts = sorted(
                            [
                                aws_account.identity_center_details.org_account_map[
                                    nested_account_id
                                ]
                                for nested_account_id in details["accounts"]
                            ]
                        )
                    account_rules[account_rule_key] = {
                        "included_orgs": [aws_account.org_id],
                        "included_accounts": accounts,
                        "users": [],
                        "groups": [],
                        "account_rule_key": account_rule_key,  # this will be remove after sorting
                    }

                account_rules[account_rule_key][f"{assignment_type}s"].append(
                    details.get("user_name", details["display_name"])
                )

        if account_rules:
            template_params["access_rules"].extend(list(account_rules.values()))

    # access_rules has no semantic ordering, we sort it explicitly to ensure the template
    # generation is deterministic.
    template_params["access_rules"] = _sorted_and_clean_access_rules(
        template_params["access_rules"]
    )

    if len(permission_set_refs) != len(aws_account_map):
        template_params["included_orgs"] = [
            aws_account_map[permission_set["account_id"]].org_id
            for permission_set in permission_set_refs
        ]

    if session_duration_resources:
        template_properties["session_duration"] = await group_int_or_str_attribute(
            aws_account_map,
            num_of_accounts,
            session_duration_resources,
            "session_duration",
        )

    if permissions_boundary_resources:
        template_properties["permissions_boundary"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            permissions_boundary_resources,
            prefer_templatized=prefer_templatized,
        )

    if description_resources:
        template_properties["description"] = await group_int_or_str_attribute(
            aws_account_map, num_of_accounts, description_resources, "description"
        )

    if managed_policy_resources:
        template_properties["managed_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            managed_policy_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if customer_managed_policy_ref_resources:
        template_properties[
            "customer_managed_policy_references"
        ] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            customer_managed_policy_ref_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if tag_resources:
        template_properties["tags"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            tag_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if inline_policy_resources:
        template_properties["inline_policy"] = json.loads(
            await group_int_or_str_attribute(  # hm... other (role, user) inline policies use group_dict_attribute, not sure the reason
                aws_account_map,
                num_of_accounts,
                inline_policy_resources,
                "inline_policy",
            )
        )

    if relay_state_resources:
        template_properties["relay_state"] = await group_int_or_str_attribute(
            aws_account_map, num_of_accounts, relay_state_resources, "relay_state"
        )

    file_path = get_templated_permission_set_file_path(
        permission_set_dir, permission_set_name
    )
    return create_or_update_template(
        file_path,
        existing_template_map,
        permission_set_name,
        AwsIdentityCenterPermissionSetTemplate,
        template_params,
        PermissionSetProperties(**template_properties),
        list(aws_account_map.values()),
    )


async def collect_aws_permission_sets(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    identity_center_template_map: dict,
    detect_messages: list[PermissionSetMessageDetails] = None,
):
    resource_dir = exe_message.get_directory(*RESOURCE_DIR)
    aws_account_map = await get_aws_account_map(config)
    if exe_message.provider_id:
        aws_account_map = {
            exe_message.provider_id: aws_account_map[exe_message.provider_id]
        }

    if detect_messages:
        detect_messages = [
            msg
            for msg in detect_messages
            if isinstance(msg, PermissionSetMessageDetails)
        ]
        if not detect_messages:
            return

    identity_center_accounts = []
    if config.accounts:
        identity_center_accounts.extend(
            [account for account in config.accounts if account.identity_center_details]
        )

    if not identity_center_accounts:
        return

    log.info("Generating AWS Identity Center Permission Set templates.")
    log.info(
        "Beginning to retrieve AWS Identity Center Permission Sets.",
        org_accounts=list(aws_account_map.keys()),
    )

    await asyncio.gather(
        *[account.set_identity_center_details() for account in identity_center_accounts]
    )

    messages = []
    permission_set_names = []
    if detect_messages:
        permission_sets_in_aws = set(
            list(
                itertools.chain.from_iterable(
                    [
                        aws_account.identity_center_details.permission_set_map.keys()
                        for aws_account in identity_center_accounts
                    ]
                )
            )
        )

        permission_set_names = await gather_permission_set_names(
            identity_center_accounts,
            detect_messages,
        )
        permission_set_names = [
            name for name in permission_set_names if name in permission_sets_in_aws
        ]

    # Upsert permission sets
    for aws_account in aws_account_map.values():
        if not aws_account.identity_center_details:
            continue

        instance_arn = aws_account.identity_center_details.instance_arn
        identity_center_client = await aws_account.get_boto3_client(
            "sso-admin", region_name=aws_account.identity_center_details.region_name
        )

        if permission_set_names:
            for name in permission_set_names:
                if permission_set := aws_account.identity_center_details.permission_set_map.get(
                    name
                ):
                    messages.append(
                        dict(
                            account_id=aws_account.account_id,
                            identity_center_client=identity_center_client,
                            instance_arn=instance_arn,
                            permission_set=permission_set,
                            user_map=aws_account.identity_center_details.user_map,
                            group_map=aws_account.identity_center_details.group_map,
                            account_resource_dir=resource_dir,
                        )
                    )
        else:
            for (
                permission_set
            ) in aws_account.identity_center_details.permission_set_map.values():
                messages.append(
                    dict(
                        account_id=aws_account.account_id,
                        identity_center_client=identity_center_client,
                        instance_arn=instance_arn,
                        permission_set=permission_set,
                        user_map=aws_account.identity_center_details.user_map,
                        group_map=aws_account.identity_center_details.group_map,
                        account_resource_dir=resource_dir,
                    )
                )

    log.info(
        "Beginning to enrich AWS IAM Identity Center Permission Sets.",
        org_accounts=list(aws_account_map.keys()),
        permission_set_count=len(messages),
    )
    generate_permission_set_resource_file_semaphore = NoqSemaphore(
        generate_permission_set_resource_file, 30
    )
    all_permission_sets = await generate_permission_set_resource_file_semaphore.process(
        messages
    )

    log.info(
        "Finished enriching AWS IAM Identity Center Permission Sets.",
        org_accounts=list(aws_account_map.keys()),
        permission_set_count=len(messages),
    )

    all_permission_sets_output = json.dumps(all_permission_sets)
    with open(
        exe_message.get_file_path(*RESOURCE_DIR, file_name_and_extension="output.json"),
        "w",
    ) as f:
        f.write(all_permission_sets_output)


async def generate_aws_permission_set_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    identity_center_template_map: dict,
    detect_messages: list[PermissionSetMessageDetails] = None,
):
    resource_dir = get_template_dir(base_output_dir)
    aws_account_map = await get_aws_account_map(config)
    existing_template_map = identity_center_template_map.get(
        AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE, {}
    )

    all_permission_sets = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR, file_name_and_extension="output.json", flatten_results=True
    )

    permission_sets_grouped_by_account = defaultdict(list)
    # list[dict(account_id:str, resources=list[dict(resource_val: str, **)])]

    for elem in range(len(all_permission_sets)):
        account_id = all_permission_sets[elem].pop("account_id")
        all_permission_sets[elem]["resource_val"] = all_permission_sets[elem][
            "permission_set"
        ]["Name"]
        permission_sets_grouped_by_account[account_id].append(all_permission_sets[elem])

    all_permission_sets = [
        dict(account_id=account_id, resources=permission_sets)
        for account_id, permission_sets in permission_sets_grouped_by_account.items()
    ]

    grouped_permission_set_map = await base_group_str_attribute(
        aws_account_map, all_permission_sets
    )

    log.info(
        "Writing templated AWS Identity Center Permission Set.",
        unique_identities=len(grouped_permission_set_map),
    )
    all_resource_ids = set()
    for name, refs in grouped_permission_set_map.items():
        resource_template = await create_templated_permission_set(
            aws_account_map,
            name,
            refs,
            resource_dir,
            existing_template_map,
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    if not detect_messages:
        # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated AWS Identity Center Permission Set generation")



from __future__ import annotations

from typing import Any, Optional

from okta.client import Client as OktaClient
from pydantic import BaseModel, Extra, Field, SecretStr, validator

from iambic.core.iambic_enum import IambicManaged
from iambic.core.iambic_plugin import ProviderPlugin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.okta.app.models import OktaAppTemplate
from iambic.plugins.v0_1_0.okta.group.models import OktaGroupTemplate
from iambic.plugins.v0_1_0.okta.handlers import import_okta_resources, load
from iambic.plugins.v0_1_0.okta.user.models import OktaUserTemplate


class OktaOrganization(BaseModel):
    idp_name: str
    org_url: str
    api_token: SecretStr
    request_timeout: int = 60
    client: Any = None  # OktaClient
    iambic_managed: Optional[IambicManaged] = Field(
        IambicManaged.UNDEFINED,
        description="Controls the directionality of iambic changes",
    )

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid

    async def get_okta_client(self) -> OktaClient:
        if not self.client:
            self.client = OktaClient(
                {
                    "orgUrl": self.org_url,
                    "token": self.api_token.get_secret_value(),
                    "requestTimeout": self.request_timeout,
                    "rateLimit": {"maxRetries": 0},
                }
            )
        return self.client


class OktaConfig(BaseModel):
    organizations: list[OktaOrganization] = Field(
        description="A list of Okta organizations."
    )

    @validator(
        "organizations", allow_reuse=True
    )  # the need of allow_reuse is possibly related to how we handle inheritance
    def validate_okta_organizations(cls, orgs: list[OktaOrganization]):
        idp_name_set = set()
        for org in orgs:
            if org.idp_name in idp_name_set:
                raise ValueError(
                    f"idp_name must be unique within organizations: {org.idp_name}"
                )
            else:
                idp_name_set.add(org.idp_name)
        return orgs

    def get_organization(self, idp_name: str) -> OktaOrganization:
        for o in self.organizations:
            if o.idp_name == idp_name:
                return o
        raise Exception(f"Could not find organization for IDP {idp_name}")


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="okta",
    version=PLUGIN_VERSION,
    provider_config=OktaConfig,
    requires_secret=True,
    async_import_callable=import_okta_resources,
    async_load_callable=load,
    templates=[
        OktaAppTemplate,
        OktaGroupTemplate,
        OktaUserTemplate,
    ],
)



from __future__ import annotations

import glob
import os

from iambic.core.logger import log


def create_workflow_files(
    repo_dir: str, repo_name: str, commit_email: str, assume_role_arn: str, region: str
):
    replacement_val_map = {
        # Leave the example here
        # "__TEMPLATE_IAMBIC_COMMIT_EMAIL": commit_email,
    }

    file_paths = glob.glob(f"{os.path.dirname(__file__)}/**/*.yaml", recursive=True)
    file_paths += glob.glob(f"{os.path.dirname(__file__)}/**/*.yml", recursive=True)

    workflows_dir = f"{str(repo_dir)}/.github/workflows"
    os.makedirs(workflows_dir, exist_ok=True)

    for file_path in file_paths:
        with open(file_path, mode="r") as f:
            file_content = f.read()
            for k, v in replacement_val_map.items():
                file_content = file_content.replace(k, v)

        with open(f"{workflows_dir}/{os.path.basename(file_path)}", mode="w") as f:
            f.write(file_content)

    log.info("Created workflow files successfully.", file_location=workflows_dir)



from __future__ import annotations

import asyncio
from itertools import chain
from typing import Callable, Optional, Union

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.iam.group.utils import (
    apply_group_inline_policies,
    apply_group_managed_policies,
    delete_iam_group,
    get_group,
)
from iambic.plugins.v0_1_0.aws.iam.models import Path
from iambic.plugins.v0_1_0.aws.iam.policy.models import ManagedPolicyRef, PolicyDocument
from iambic.plugins.v0_1_0.aws.models import AccessModel, AWSAccount, AWSTemplate
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call

AWS_IAM_GROUP_TEMPLATE_TYPE = "NOQ::AWS::IAM::Group"


class GroupProperties(BaseModel):
    group_name: str = Field(
        description="Name of the group",
    )
    path: Optional[Union[str, list[Path]]] = "/"
    managed_policies: Optional[list[ManagedPolicyRef]] = Field(
        [],
        description="Managed policy arns attached to the group",
    )
    inline_policies: Optional[list[PolicyDocument]] = Field(
        [],
        description="List of the group's inline policies",
    )

    @property
    def resource_type(self):
        return "aws:iam:group"

    @property
    def resource_id(self):
        return self.group_name

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("managed_policies")
    def sort_managed_policy_refs(cls, v: list[ManagedPolicyRef]):
        sorted_v = sorted(v, key=cls.sort_func("policy_arn"))
        return sorted_v

    @validator("inline_policies")
    def sort_inline_policies(cls, v: list[PolicyDocument]):
        sorted_v = sorted(v, key=cls.sort_func("policy_name"))
        return sorted_v

    @validator("path")
    def sort_path(cls, v: list[Path]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v


class AwsIamGroupTemplate(AWSTemplate, AccessModel):
    template_type = AWS_IAM_GROUP_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the group")
    properties: GroupProperties = Field(
        description="Properties of the group",
    )

    def _is_iambic_import_only(self):
        return "aws-service-group" in self.properties.path

    async def _apply_to_account(  # noqa: C901
        self, aws_account: AWSAccount
    ) -> AccountChangeDetails:
        client = await aws_account.get_boto3_client("iam")
        account_group = self.apply_resource_dict(aws_account)

        group_name = account_group["GroupName"]
        account_change_details = AccountChangeDetails(
            account=str(aws_account),
            resource_id=group_name,
            resource_type=self.resource_type,
            new_value=dict(**account_group),
            proposed_changes=[],
            exceptions_seen=[],
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_id=group_name,
            account=str(aws_account),
        )
        if self._is_iambic_import_only():
            return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        current_group = await get_group(
            group_name, client, include_policies=bool(not deleted)
        )
        if current_group:
            account_change_details.current_value = {
                **current_group
            }  # Create a new dict

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        if deleted:
            if current_group:
                account_change_details.new_value = None
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=group_name,
                        resource_type=self.resource_type,
                    )
                ]
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                if ctx.execute:
                    apply_awaitable = delete_iam_group(group_name, client, log_params)
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        group_exists = bool(current_group)
        inline_policies = account_group.pop("InlinePolicies", [])
        managed_policies = account_group.pop("ManagedPolicies", [])
        existing_inline_policies = current_group.pop("InlinePolicies", [])
        existing_managed_policies = current_group.pop("ManagedPolicies", [])
        tasks = []

        if group_exists:
            tasks.extend([])

            supported_update_keys = ["Path", "GroupName"]
            update_resource_log_params = {**log_params}
            update_group_params = {}
            for k in supported_update_keys:
                if account_group.get(k) is not None and account_group.get(
                    k
                ) != current_group.get(k):
                    update_resource_log_params[k] = dict(
                        old_value=current_group.get(k),
                        new_value=account_group.get(k),
                    )
                    update_group_params[f"New{k}"] = account_group.get(k)

            if update_group_params:
                log_str = "Out of date resource found."
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.UPDATE,
                        resource_id=group_name,
                        resource_type=self.resource_type,
                    )
                ]
                if ctx.execute:
                    log.debug(
                        f"{log_str} Updating resource...",
                        **update_resource_log_params,
                    )
                    apply_awaitable = boto_crud_call(
                        client.update_group,
                        RoleName=group_name,
                        **update_group_params,
                    )
                    tasks.append(
                        plugin_apply_wrapper(apply_awaitable, proposed_changes)
                    )
                else:
                    log.debug(log_str, **update_resource_log_params)
                    account_change_details.extend_changes(proposed_changes)
        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=group_name,
                    resource_type=self.resource_type,
                )
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log.debug(f"{log_str} Creating resource...", **log_params)
            apply_awaitable = boto_crud_call(client.create_group, **account_group)
            account_change_details.extend_changes(
                await plugin_apply_wrapper(apply_awaitable, proposed_changes)
            )

            if account_change_details.exceptions_seen:
                log.error(
                    "Unable to create resource on account",
                    exceptions_seen=[
                        cd.exceptions_seen
                        for cd in account_change_details.exceptions_seen
                    ],
                    **log_params,
                )
                return account_change_details

        tasks.extend(
            [
                apply_group_managed_policies(
                    group_name,
                    client,
                    managed_policies,
                    existing_managed_policies,
                    log_params,
                ),
                apply_group_inline_policies(
                    group_name,
                    client,
                    inline_policies,
                    existing_inline_policies,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            account_change_details.extend_changes(
                list(chain.from_iterable(changes_made))
            )

        if ctx.execute and not account_change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

        return account_change_details



from __future__ import annotations

import json
from datetime import datetime
from decimal import Decimal
from typing import Any
from uuid import UUID

import ujson
from deepdiff.model import PrettyOrderedSet


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (frozenset, set, PrettyOrderedSet)):
            return list(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S %Z")
        if isinstance(obj, Exception):
            return str(obj)
        if isinstance(obj, UUID):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def dumps(
    obj: any,
    ensure_ascii: bool = True,
    double_precision: int = 9,
    encode_html_chars: bool = False,
    escape_forward_slashes: bool = False,
    sort_keys: bool = False,
    indent: int = 0,
    **kwargs,
) -> str:
    # Try fast "ujson is 3x faster than the standard json library" first
    try:
        result = ujson.dumps(
            obj,
            ensure_ascii=ensure_ascii,
            double_precision=double_precision,
            encode_html_chars=encode_html_chars,
            escape_forward_slashes=escape_forward_slashes,
            sort_keys=sort_keys,
            indent=indent,
            **kwargs,
        )
    except TypeError:
        result = json.dumps(
            obj,
            cls=SetEncoder,
            ensure_ascii=ensure_ascii,
            sort_keys=sort_keys,
            indent=indent,
            **kwargs,
        )
    return result


def loads(s: str, **kwargs) -> Any:
    try:
        result = ujson.loads(s, **kwargs)
    except ValueError:
        result = json.loads(s, **kwargs)
    return result



from __future__ import annotations

from typing import List

from iambic.core.models import TemplateChangeDetails
from iambic.output import get_template_env
from iambic.output.models import get_template_data


def gh_render_resource_changes(resource_changes: List[TemplateChangeDetails]):
    template_data = get_template_data(resource_changes)
    env = get_template_env()
    template = env.get_template("github_summary.jinja2")
    return template.render(iambic=template_data)



from __future__ import annotations

from pydantic import BaseModel, Field, validator

from iambic.core.iambic_plugin import ProviderPlugin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.azure_ad.group.models import (
    AzureActiveDirectoryGroupTemplate,
)
from iambic.plugins.v0_1_0.azure_ad.handlers import import_azure_ad_resources, load
from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization
from iambic.plugins.v0_1_0.azure_ad.user.models import AzureActiveDirectoryUserTemplate


class AzureADConfig(BaseModel):
    organizations: list[AzureADOrganization] = Field(
        description="A list of Azure Active Directory organizations."
    )

    @validator(
        "organizations", allow_reuse=True
    )  # the need of allow_reuse is possibly related to how we handle inheritance
    def validate_azure_ad_organizations(cls, orgs: list[organizations]):
        idp_name_set = set()
        for org in orgs:
            if org.idp_name in idp_name_set:
                raise ValueError(
                    f"idp_name must be unique within organizations: {org.idp_name}"
                )
            else:
                idp_name_set.add(org.idp_name)
        return orgs

    def get_organization(self, idp_name: str) -> AzureADOrganization:
        for o in self.organizations:
            if o.idp_name == idp_name:
                return o
        raise Exception(f"Could not find organization for IDP {idp_name}")


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="azure_ad",
    version=PLUGIN_VERSION,
    provider_config=AzureADConfig,
    async_import_callable=import_azure_ad_resources,
    async_load_callable=load,
    requires_secret=True,
    templates=[AzureActiveDirectoryGroupTemplate, AzureActiveDirectoryUserTemplate],
)



from __future__ import annotations

import json
import os
import sys
import time
import traceback
from functools import partial
from typing import Union

from pydantic import ValidationError
from ruamel.yaml.scanner import ScannerError

from iambic.config.templates import TEMPLATES
from iambic.core.logger import log
from iambic.core.models import BaseTemplate
from iambic.core.utils import transform_comments, yaml

# we must avoid import multiprocessing pool in the module loading time
if os.environ.get("AWS_LAMBDA_FUNCTION_NAME", False):
    from multiprocessing import cpu_count

    from iambic.vendor.lambda_multiprocessing import Pool
else:
    from multiprocessing import Pool, cpu_count


# line number is zero-th based
def resolve_location(loc_list: list[str], ruamel_dict) -> Union[None, int]:
    local_loc_list = loc_list
    local_ruamel_dict = ruamel_dict
    last_known_location = None
    if len(local_loc_list) == 0:
        return None
    while len(local_loc_list) > 1:
        lookup_key = local_loc_list[0]
        peek_ruamel_dict = ruamel_dict.get(str.lower(lookup_key), None)
        if peek_ruamel_dict is None:
            return None
        else:
            last_known_location = local_ruamel_dict.lc.data.get(str.lower(lookup_key))
        local_ruamel_dict = ruamel_dict.get(str.lower(lookup_key), None)
        local_loc_list = loc_list[1:]
    line_info = local_ruamel_dict.lc.data.get(str.lower(local_loc_list[0]), None)
    if line_info is None:
        if last_known_location is not None:
            return last_known_location[0]
        else:
            return None
    else:
        return line_info[0]


def format_validation_error(err, ruamel_dict):
    try:
        errors = json.loads(err.json())
        lines = []
        for error in errors:
            if error["type"] == "value_error.missing":
                line_num = resolve_location(error["loc"], ruamel_dict)
                canonical_key = str.lower(".".join(error["loc"]))
                if line_num is not None:
                    missing_key = str.lower(error["loc"][-1])
                    lines.append(
                        f"Missing Field: `{missing_key}` around line {line_num+1}"
                    )
                else:
                    lines.append(f"Missing Field: `{canonical_key}`")
            if error["type"].startswith("type_error"):
                line_num = resolve_location(error["loc"], ruamel_dict)
                canonical_key = str.lower(".".join(error["loc"]))
                lines.append(f"line {line_num+1}: `{canonical_key}` has type issue")
        return "\n".join(lines)
    except Exception:
        # need to still return something to avoid to downstream formatting
        captured_traceback = traceback.format_exc()
        return f"Unable to compute hints: {captured_traceback}"


def load_template(template_path: str, raise_validation_err: bool = True) -> dict:
    try:
        template_dict = transform_comments(yaml.load(open(template_path)))
        template_type = template_dict.get("template_type")
        if template_type and template_type not in ["NOQ::Core::Config"]:
            template_dict["file_path"] = template_path
            return template_dict
    except ScannerError as err:
        log.critical(
            "Invalid template structure", file_path=template_path, error=repr(err)
        )
        if raise_validation_err:
            raise ValueError(f"{template_path} template has validation error.") from err


def load_templates(
    template_paths: list[str],
    raise_validation_err: bool = True,
    use_multiprocessing=True,
) -> list[BaseTemplate]:
    templates = []
    template_dicts = None

    if use_multiprocessing:
        load_template_fn = partial(
            load_template, raise_validation_err=raise_validation_err
        )
        with Pool(max(1, cpu_count() // 2)) as p:
            template_dicts = p.map(load_template_fn, template_paths)
            if getattr(sys, "gettrace", None):
                # When in debug mode, subprocesses can exit
                # before debugger can attach
                # https://github.com/microsoft/debugpy/issues/712
                time.sleep(0.5)
    else:
        template_dicts = [
            load_template(path, raise_validation_err) for path in template_paths
        ]

    for template_dict in template_dicts:
        if not template_dict:
            continue

        try:
            template_cls = TEMPLATES.template_map[template_dict["template_type"]]
            template_cls.update_forward_refs()
            templates.append(template_cls(**template_dict))
        except KeyError:
            log.critical(
                "Invalid template type",
                file_path=template_dict["file_path"],
                template_type=template_dict["template_type"],
            )
            # We should allow to continue to allow unknown template type; otherwise,
            # we cannot support forward or backward compatibility during version changes.
        except ValidationError as err:
            log.critical(
                "Invalid template structure",
                file_path=template_dict["file_path"],
                error=repr(err),
            )
            if raise_validation_err:
                hints = format_validation_error(err, template_dict)
                raise ValueError(
                    f"{template_dict['file_path']} template has validation error. \n{hints}"
                ) from err

    return templates



from __future__ import annotations

import asyncio
from typing import Any, Optional

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from iambic.core.models import BaseTemplate, ExecutionMessage, TemplateChangeDetails


async def default_apply_callable(
    exe_message: ExecutionMessage,
    config,
    templates: list[BaseTemplate],
    remote_worker=None,
) -> list[TemplateChangeDetails]:
    """
    The default apply callable for the IambicPlugin class.

    :param exe_message: Execution context
    :param config: The plugin's config object.
    :param templates: The list of templates to apply.
    :param remote_worker: The remote worker to use for applying templates.
    """
    template_changes = await asyncio.gather(
        *[template.apply(config) for template in templates]
    )

    return [
        template_change
        for template_change in template_changes
        if template_change.proposed_changes or template_change.exceptions_seen
    ]


class ProviderPlugin(PydanticBaseModel):
    version: str = Field(description="The version of the plugin.")
    config_name: str = Field(
        description="The name of the provider configuration in the iambic config file."
    )
    requires_secret: bool = Field(
        default=False,
        description="Whether or not the provider requires a secret to be passed in.",
    )
    provider_config: Any = Field(
        description="The Pydantic model that is attached to the Config."
        "This will contain the provider specific configuration."
        "These are things like the AWSAccount model, OktaOrganization or GoogleProject."
    )
    async_load_callable: Any = Field(
        description="The function that is called to load any dynamic metadata used by the provider."
        "For example, assigning default session info to an AWS account or decoding a secret."
        "This function must accept the param (config: ProviderConfig)."
        "The changes must be made to the config object directly and must return the config",
        hidden_from_schema=True,
    )
    async_import_callable: Any = Field(
        description="The function that called to import resources across all templates for this provider."
        "This function must accept the "
        "params: (exe_message: ExecutionMessage, config: ProviderConfig, base_output_dir: str, detect_messages: list = None, remote_worker: Worker = None)",
        hidden_from_schema=True,
    )
    async_apply_callable: Any = Field(
        description="The function that called to apply resources across all templates for this provider."
        "This function must accept the "
        "params: (exe_message: ExecutionMessage, config: ProviderConfig, templates: list[BaseTemplate], remote_worker: Worker = None)."
        "It must return a list[TemplateChangeDetails].",
        default=default_apply_callable,
        hidden_from_schema=True,
    )
    async_detect_changes_callable: Optional[Any] = Field(
        description="(OPTIONAL) The function that called to detect changes across all templates for this provider."
        "This is optional and if not provided will fallback to the async_import_callable."
        "The function is called more frequently than the import_callable."
        "It is used as a drift detection tool."
        "For example, the default AWS plugin supports an SQS queue containing cloudtrail events."
        "This function must accept the "
        "params: (config: ProviderConfig, repo_dir: str)"
        "It must return a str containing the detected changes.",
        hidden_from_schema=True,
    )
    async_decode_secret_callable: Optional[Any] = Field(
        description="(OPTIONAL) The function that called to decode a secret."
        "Check extend.key before attempting to decode."
        "This function must accept the params (config: ProviderConfig, extend: ExtendsConfig)"
        "It must return the decoded secret as a dict.",
        hidden_from_schema=True,
    )
    async_discover_upstream_config_changes_callable: Optional[Any] = Field(
        description="(OPTIONAL) The function that called to discover upstream config changes."
        "An example of this would be a new account being added to an AWS Organization,"
        "or a change to AWS account's name or tags."
        "This function must accept the params: (exe_message: ExecutionMessage, config: ProviderConfig, repo_dir: str, remote_worker: Worker = None)",
        hidden_from_schema=True,
    )
    templates: list = Field(description="The list of templates used for this provider.")



from __future__ import annotations

import os
import typing
from typing import TYPE_CHECKING, Any, Optional

import googleapiclient.discovery
from google.oauth2 import service_account
from pydantic import BaseModel, Field, SecretStr, validator

from iambic.core.iambic_enum import IambicManaged
from iambic.core.iambic_plugin import ProviderPlugin
from iambic.core.models import Variable
from iambic.core.utils import aio_wrapper
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.google_workspace.group.models import (
    GoogleWorkspaceGroupTemplate,
)
from iambic.plugins.v0_1_0.google_workspace.handlers import (
    import_google_resources,
    load,
)

if TYPE_CHECKING:  # pragma: no cover
    MappingIntStrAny = typing.Mapping[int | str, Any]
    AbstractSetIntStr = typing.AbstractSet[int | str]


class GoogleSubject(BaseModel):
    domain: str
    service_account: str


class GoogleProject(BaseModel):
    project_id: str
    project_name: Optional[str]
    subjects: list[GoogleSubject]
    type: str
    private_key_id: str
    private_key: SecretStr
    client_email: str
    client_id: str
    auth_uri: str
    token_uri: str
    auth_provider_x509_cert_url: str
    client_x509_cert_url: str
    variables: Optional[list[Variable]] = Field(
        [],
        description="A list of variables to be used when creating templates",
    )
    iambic_managed: Optional[IambicManaged] = Field(
        IambicManaged.UNDEFINED,
        description="Controls the directionality of iambic changes",
    )
    _service_connection_map: dict = {}

    def __str__(self):
        if self.project_name:
            return f"{self.project_name} - ({self.project_id})"

        return self.project_id

    async def get_service_connection(
        self,
        service_name: str,
        service_path: str,
        domain: str,
        cache_discovery: bool = False,
    ):
        # sourcery skip: raise-specific-error
        key = f"{domain}:{service_name}:{service_path}"
        if self._service_connection_map.get(key):
            return self._service_connection_map[key]
        if not os.environ.get("TESTING"):
            if service_conn := self._service_connection_map.get(key):
                return service_conn
        credentials = self.dict(
            include={
                "type",
                "project_id",
                "private_key_id",
                "private_key",
                "client_email",
                "client_id",
                "auth_uri",
                "token_uri",
                "auth_provider_x509_cert_url",
                "client_x509_cert_url",
            }
        )
        credentials["private_key"] = credentials["private_key"].get_secret_value()
        admin_credentials = service_account.Credentials.from_service_account_info(
            credentials,
            scopes=[
                "https://www.googleapis.com/auth/admin.directory.user.security",
                "https://www.googleapis.com/auth/admin.reports.audit.readonly",
                "https://www.googleapis.com/auth/admin.directory.user",
                "https://www.googleapis.com/auth/admin.directory.group",
                "https://www.googleapis.com/auth/admin.directory.group.member",
            ],
        )

        admin_delegated_credentials = None

        for s in self.subjects:
            if s.domain == domain:
                admin_delegated_credentials = admin_credentials.with_subject(
                    s.service_account
                )
                break
        if not admin_delegated_credentials:
            raise Exception(f"Could not find service account for domain {domain}")

        self._service_connection_map[key] = await aio_wrapper(
            googleapiclient.discovery.build,
            service_name,
            service_path,
            credentials=admin_delegated_credentials,
            cache_discovery=cache_discovery,
            thread_sensitive=True,
        )
        return self._service_connection_map[key]


class GoogleWorkspaceConfig(BaseModel):
    workspaces: list[GoogleProject]

    @validator(
        "workspaces", allow_reuse=True
    )  # the need of allow_reuse is possibly related to how we handle inheritance
    def validate_google_workspaces(cls, workspaces: list[GoogleProject]):
        project_id_set = set()
        for workspace in workspaces:
            if workspace.project_id in project_id_set:
                raise ValueError(
                    f"project_id must be unique within workspaces: {workspace.project_id}"
                )
            else:
                project_id_set.add(workspace.project_id)
        return workspaces

    def get_workspace(self, project_id: str) -> GoogleProject:
        for w in self.workspaces:
            if w.project_id == project_id:
                return w
        raise Exception(f"Could not find workspace for project_id {project_id}")


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="google_workspace",
    version=PLUGIN_VERSION,
    provider_config=GoogleWorkspaceConfig,
    requires_secret=True,
    async_import_callable=import_google_resources,
    async_load_callable=load,
    templates=[
        GoogleWorkspaceGroupTemplate,
    ],
)



from __future__ import annotations

from enum import Enum

from ruamel.yaml import YAML, yaml_object

yaml = YAML()


@yaml_object(yaml)
class IambicManaged(Enum):
    UNDEFINED = "undefined"
    READ_AND_WRITE = "read_and_write"
    IMPORT_ONLY = "import_only"
    ENFORCED = "enforced"
    DISABLED = "disabled"

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar("!IambicManaged", f"{node._value_}")

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.value)


class Command(Enum):
    IMPORT = "import"
    APPLY = "apply"
    CONFIG_DISCOVERY = "config_discovery"


class ExecutionStatus(Enum):
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"



from __future__ import annotations

import os
import random
from multiprocessing import Pipe, Process, TimeoutError
from multiprocessing.connection import Connection
from time import time
from typing import Any, Dict, Iterable, List, Tuple, Union
from uuid import UUID, uuid4


class Child:
    proc: Process

    # this is the number of items that we have sent to the child process
    # minus the number we have received back
    # includes items in the queue not processed,
    # items currently being processed
    # and items that have been processed by the child, but not read by the parent
    # does not include the termination command from parent to child
    queue_sz: int = 0

    # parent_conn.send()  to give stuff to the child
    # parent_conn.recv() to get results back from child
    parent_conn: Connection
    child_conn: Connection

    result_cache: Dict[UUID, Tuple[Any, Exception]] = {}

    _closed: bool = False

    # if True, do the work in the main process
    # but present the same interface
    # and still send stuff through the pipes (to verify they're pickleable)
    # this is so we can unit test with moto
    main_proc: bool

    def __init__(self, main_proc=False):
        self.parent_conn, self.child_conn = Pipe(duplex=True)
        self.main_proc = main_proc
        if not main_proc:
            self.proc = Process(target=self.spin)
            self.proc.start()

    # each child process runs in this
    # a while loop waiting for payloads from the self.child_conn
    # [(id, func, args, kwds), None] -> call func(args, *kwds)
    #                         and send the return back through the self.child_conn pipe
    #                         {id: (ret, None)} if func returned ret
    #                         {id: (None, err)} if func raised exception err
    # [None, True] -> exit gracefully (write nothing to the pipe)
    def spin(self) -> None:
        while True:
            (job, quit_signal) = self.child_conn.recv()
            if quit_signal:
                break
            else:
                (id, func, args, kwds) = job
                result = self._do_work(id, func, args, kwds)
                self.child_conn.send(result)
        self.child_conn.close()

    def _do_work(
        self, id, func, args, kwds
    ) -> Union[Tuple[Any, None], Tuple[None, Exception]]:
        try:
            ret = {id: (func(*args, **kwds), None)}
        except Exception as e:
            # how to handle KeyboardInterrupt?
            ret = {id: (None, e)}
        assert isinstance(list(ret.keys())[0], UUID)
        return ret

    def submit(self, func, args=(), kwds=None) -> "AsyncResult":
        if self._closed:
            raise ValueError("Cannot submit tasks after closure")
        if kwds is None:
            kwds = {}
        id = uuid4()
        self.parent_conn.send([(id, func, args, kwds), None])
        if self.main_proc:
            self.child_conn.recv()
            ret = self._do_work(id, func, args, kwds)
            self.child_conn.send(ret)
        self.queue_sz += 1
        return AsyncResult(id=id, child=self)

    # grab all results in the pipe from child to parent
    # save them to self.result_cache
    def flush(self):
        # watch out, when the other end is closed, a termination byte appears, so .poll() returns True
        while (
            (not self.parent_conn.closed)
            and (self.queue_sz > 0)
            and self.parent_conn.poll(0)
        ):
            result = self.parent_conn.recv()
            assert isinstance(list(result.keys())[0], UUID)
            self.result_cache.update(result)
            self.queue_sz -= 1

    # prevent new tasks from being submitted
    # but keep existing tasks running
    # should be idempotent
    def close(self):
        if not self._closed:
            if not self.main_proc:
                # send quit signal to child
                self.parent_conn.send([None, True])
            else:
                # no child process to close
                self.flush()
                self.child_conn.close()

            # keep track of closure,
            # so subsequent task submissions are rejected
            self._closed = True

    # after closing
    # wait for existing tasks to finish
    # should be idempotent
    def join(self):
        assert self._closed, "Must close before joining"
        if not self.main_proc:
            try:
                self.proc.join()
            except ValueError:
                # .join() has probably been called multiple times
                # so the process has already been closed
                pass
            finally:
                self.proc.close()

        self.flush()
        self.parent_conn.close()

    # terminate child processes without waiting for them to finish
    # should be idempotent
    def terminate(self):
        if not self.main_proc:
            try:
                a = self.proc.is_alive()
            except ValueError:
                # already closed
                # .is_alive seems to raise ValueError not return False if dead
                pass
            else:
                if a:
                    try:
                        self.proc.close()
                    except ValueError:
                        self.proc.terminate()
        self.parent_conn.close()
        self.child_conn.close()
        self._closed |= True

    def __del__(self):
        self.terminate()


class AsyncResult:
    def __init__(self, id: UUID, child: Child):
        assert isinstance(id, UUID)
        self.id = id
        self.child = child
        self.result: Union[Tuple[Any, None], Tuple[None, Exception]] = None

    # assume the result is in the self.child.result_cache
    # move it into self.result
    def _load(self):
        self.result = self.child.result_cache[self.id]
        del self.child.result_cache[self.id]  # prevent memory leak

    # Return the result when it arrives.
    # If timeout is not None and the result does not arrive within timeout seconds
    # then multiprocessing.TimeoutError is raised.
    # If the remote call raised an exception then that exception will be reraised by get().
    # .get() must remember the result
    # and return it again multiple times
    # delete it from the Child.result_cache to avoid memory leak
    def get(self, timeout=None):
        if self.result is not None:
            (response, ex) = self.result
            if ex:
                raise ex
            else:
                return response
        elif self.id in self.child.result_cache:
            self._load()
            return self.get(0)
        else:
            self.wait(timeout)
            if not self.ready():
                raise TimeoutError("result not ready")
            else:
                return self.get(0)

    # Wait until the result is available or until timeout seconds pass.
    def wait(self, timeout=None):
        start_t = time()
        if self.result is None:
            self.child.flush()
            # the result we want might not be the next result
            # it might be the 2nd or 3rd next
            while (self.id not in self.child.result_cache) and (
                (timeout is None) or (time() - timeout < start_t)
            ):
                if timeout is None:
                    self.child.parent_conn.poll()
                else:
                    elapsed_so_far = time() - start_t
                    remaining = timeout - elapsed_so_far
                    self.child.parent_conn.poll(remaining)
                if self.child.parent_conn.poll(0):
                    self.child.flush()

    # Return whether the call has completed.
    def ready(self):
        self.child.flush()
        return self.result or (self.id in self.child.result_cache)

    # Return whether the call completed without raising an exception.
    # Will raise ValueError if the result is not ready.
    def successful(self):
        if self.result is None:
            if not self.ready():
                raise ValueError("Result is not ready")
            else:
                self._load()

        return self.result[1] is None


class Pool:
    def __init__(
        self,
        processes=None,
        initializer=None,
        initargs=None,
        maxtasksperchild=None,
        context=None,
    ):
        if processes is None:
            self.num_processes = os.cpu_count()
        else:
            if processes < 0:
                raise ValueError("processes must be a positive integer")
            self.num_processes = processes

        if initializer:
            raise NotImplementedError("initializer not implemented")

        if initargs:
            raise NotImplementedError("initargs not implemented")

        if maxtasksperchild:
            raise NotImplementedError("maxtasksperchild not implemented")

        if context:
            raise NotImplementedError("context not implemented")

        self._closed = False

    def __enter__(self):
        self._closed = False

        if self.num_processes > 0:
            self.children = [Child() for _ in range(self.num_processes)]
        else:
            # create one 'child' which will just do work in the main thread
            self.children = [Child(main_proc=True)]

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self.join()
        self.terminate()

    def __del__(self):
        self.terminate()

    # prevent new tasks from being submitted
    # but keep existing tasks running
    def close(self):
        if not self._closed:
            for c in self.children:
                c.close()
            self._closed |= True

    # wait for existing tasks to finish
    def join(self):
        assert self._closed, "Must close before joining"
        for c in self.children:
            c.join()

    # terminate child processes without waiting for them to finish
    def terminate(self):
        for c in self.children:
            c.terminate()
        self._closed |= True

    def apply(self, func, args=(), kwds=None):
        ret = self.apply_async(func, args, kwds)
        return ret.get()

    def apply_async(
        self, func, args=(), kwds=None, callback=None, error_callback=None
    ) -> AsyncResult:
        if callback:
            raise NotImplementedError("callback not implemented")
        if error_callback:
            raise NotImplementedError("error_callback not implemented")

        if self._closed:
            raise ValueError("Pool already closed")
        if kwds is None:
            kwds = {}

        # choose the first idle process if there is one
        # if not, choose the process with the shortest queue
        for c in self.children:
            c.flush()
        min_q_sz = min(c.queue_sz for c in self.children)
        c = random.choice([c for c in self.children if c.queue_sz <= min_q_sz])
        return c.submit(func, args, kwds)

    def map_async(
        self, func, iterable, chunksize=None, callback=None, error_callback=None
    ) -> List[AsyncResult]:
        return self.starmap_async(
            func, zip(iterable), chunksize, callback, error_callback
        )

    def map(
        self, func, iterable, chunksize=None, callback=None, error_callback=None
    ) -> List:
        return self.starmap(func, zip(iterable), chunksize, callback, error_callback)

    def starmap_async(
        self,
        func,
        iterable: Iterable[Iterable],
        chunksize=None,
        callback=None,
        error_callback=None,
    ) -> List[AsyncResult]:
        if chunksize:
            raise NotImplementedError(
                "Haven't implemented chunksizes. Infinite chunksize only."
            )
        if callback or error_callback:
            raise NotImplementedError("Haven't implemented callbacks")
        return [self.apply_async(func, args) for args in iterable]

    def starmap(
        self,
        func,
        iterable: Iterable[Iterable],
        chunksize=None,
        callback=None,
        error_callback=None,
    ) -> List[Any]:
        results = self.starmap_async(
            func, iterable, chunksize, callback, error_callback
        )
        return [r.get() for r in results]

    def imap(self, func, iterable, chunksize=None):
        raise NotImplementedError(
            "Only normal apply, map, starmap and their async equivilents have been implemented"
        )

    def imap_unordered(self, func, iterable, chunksize=None):
        raise NotImplementedError(
            "Only normal apply, map, starmap and their async equivilents have been implemented"
        )



from __future__ import annotations

import asyncio
import os
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Union

import boto3
import botocore
from aws_error_utils.aws_error_utils import errors
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra, Field, constr, validator
from ruamel.yaml import YAML, yaml_object

from iambic.core.context import ctx
from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    AccessModelMixin,
    AccountChangeDetails,
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    ProviderChild,
    TemplateChangeDetails,
    Variable,
)
from iambic.core.utils import (
    NoqSemaphore,
    evaluate_on_provider,
    get_provider_value,
    sort_dict,
)
from iambic.plugins.v0_1_0.aws.utils import (
    RegionName,
    boto_crud_call,
    create_assume_role_session,
    get_current_role_arn,
    legacy_paginated_search,
    set_org_account_variables,
)

yaml = YAML()

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

ARN_RE = r"(^arn:([^:]*):([^:]*):([^:]*):(|\*|[\d]{12}|cloudfront|aws|{{var.account_id}}):(.+)$)|^\*$"

IAMBIC_HUB_ROLE_NAME = "IambicHubRole"
IAMBIC_SPOKE_ROLE_NAME = "IambicSpokeRole"


def get_hub_role_arn(account_id: str) -> str:
    return f"arn:aws:iam::{account_id}:role/{IAMBIC_HUB_ROLE_NAME}"


def get_spoke_role_arn(account_id: str) -> str:
    return f"arn:aws:iam::{account_id}:role/{IAMBIC_SPOKE_ROLE_NAME}"


@yaml_object(yaml)
class Partition(Enum):
    AWS = "aws"
    AWS_GOV = "aws-us-gov"
    AWS_CHINA = "aws-cn"

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar("!Partition", f"{node._value_}")

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.value)


class AccessModel(AccessModelMixin, BaseModel):
    included_accounts: list[str] = Field(
        ["*"],
        description=(
            "A list of account ids and/or account names this statement applies to. "
            "Account ids/names can be represented as a regex and string"
        ),
    )
    excluded_accounts: Optional[list[str]] = Field(
        [],
        description=(
            "A list of account ids and/or account names this statement explicitly does not apply to. "
            "Account ids/names can be represented as a regex and string"
        ),
    )
    included_orgs: list[str] = Field(
        ["*"],
        description=(
            "A list of AWS organization ids this statement applies to. "
            "Org ids can be represented as a regex and string"
        ),
    )
    excluded_orgs: Optional[list[str]] = Field(
        [],
        description=(
            "A list of AWS organization ids this statement explicitly does not apply to. "
            "Org ids can be represented as a regex and string"
        ),
    )

    @validator(
        "included_accounts", "excluded_accounts", "included_orgs", "excluded_orgs"
    )
    def sort_list_of_str(cls, v: list[str]):
        return sorted(v)

    @property
    def included_children(self):
        return self.included_accounts

    def set_included_children(self, value):
        self.included_accounts = value

    @property
    def excluded_children(self):
        return self.excluded_accounts

    def set_excluded_children(self, value):
        self.excluded_accounts = value

    @property
    def included_parents(self):
        return self.included_orgs

    def set_included_parents(self, value):
        self.included_orgs = value

    @property
    def excluded_parents(self):
        return self.excluded_orgs

    def set_excluded_parents(self, value):
        self.excluded_orgs = value


class Tag(ExpiryModel, AccessModel):
    key: str
    value: str

    @property
    def resource_type(self):
        return "Tag"

    @property
    def resource_id(self):
        return f"{self.key}:{self.value}"


class BaseAWSAccountAndOrgModel(PydanticBaseModel):
    default_region: Optional[RegionName] = Field(
        RegionName.us_east_1,
        description="Default region to use when making AWS requests",
    )
    aws_profile: Optional[str] = Field(
        None,
        description="The AWS profile used when making calls to the account",
    )
    hub_role_arn: Optional[str] = Field(
        None,
        description="The role arn to assume into when making calls to the account",
    )
    external_id: Optional[str] = Field(
        None,
        description="The external id to use for assuming into a role when making calls to the account",
    )
    boto3_session_map: Optional[dict] = None

    class Config:
        fields = {"boto3_session_map": {"exclude": True}}
        extra = Extra.forbid

    @property
    def region_name(self):
        return (
            self.default_region
            if isinstance(self.default_region, str)
            else self.default_region.value
        )

    async def get_boto3_session(self, region_name: Optional[str] = None):
        region_name = region_name or self.region_name

        if self.boto3_session_map is None:
            self.boto3_session_map = {}
        elif boto3_session := self.boto3_session_map.get(region_name):
            return boto3_session

        session = boto3.Session(region_name=region_name)
        if self.aws_profile:
            try:
                session = boto3.Session(
                    profile_name=self.aws_profile, region_name=region_name
                )
            except Exception as err:
                log.warning(err)

        sts_client = session.client("sts", region_name=region_name)
        if self.hub_role_arn and self.hub_role_arn != get_current_role_arn(sts_client):
            boto3_session = await create_assume_role_session(
                session,
                self.hub_role_arn,
                region_name,
                external_id=self.external_id,
                session_name=os.environ.get("IAMBIC_SESSION_NAME", None),
            )
            if boto3_session:
                self.boto3_session_map[region_name] = boto3_session
                return boto3_session

        self.boto3_session_map[region_name] = session
        return self.boto3_session_map[region_name]

    async def get_boto3_client(self, service: str, region_name: Optional[str] = None):
        region_name = region_name or self.region_name

        if self.boto3_session_map is None:
            self.boto3_session_map = {}

        if (
            client := self.boto3_session_map.get("client", {})
            .get(service, {})
            .get(region_name)
        ):
            return client

        client = (await self.get_boto3_session(region_name)).client(
            service,
            config=botocore.client.Config(
                max_pool_connections=50, region_name=region_name
            ),
        )
        self.boto3_session_map.setdefault("client", {}).setdefault(service, {})[
            region_name
        ] = client
        return client

    async def get_active_regions(self) -> list[str]:
        client = await self.get_boto3_client("ec2")
        res = await boto_crud_call(client.describe_regions)
        return [region["RegionName"] for region in res["Regions"]]


class IdentityCenterDetails(PydanticBaseModel):
    region: RegionName = Field(RegionName.us_east_1)
    instance_arn: Optional[str] = None
    identity_store_id: Optional[str] = None
    permission_set_map: Optional[Union[dict, None]] = None
    user_map: Optional[Union[dict, None]] = None
    group_map: Optional[Union[dict, None]] = None
    org_account_map: Optional[Union[dict, None]] = None

    @property
    def region_name(self):
        return self.region if isinstance(self.region, str) else self.region.value

    def get_resource_name(self, resource_type: str, resource_id: str):
        resource_type = resource_type.lower()
        name_map = {"user": "UserName", "group": "DisplayName"}
        resource_map = getattr(self, f"{resource_type.lower()}_map")
        return resource_map.get(resource_id)[name_map[resource_type]]


class AWSAccount(ProviderChild, BaseAWSAccountAndOrgModel):
    account_id: constr(min_length=12, max_length=12) = Field(
        None, description="The AWS Account ID"
    )
    org_id: Optional[str] = Field(
        None,
        description="A unique identifier designating the identity of the organization",
    )
    account_name: str
    partition: Optional[Partition] = Field(
        Partition.AWS,
        description="The AWS partition the account is in. Options are aws, aws-us-gov, and aws-cn",
    )
    variables: Optional[List[Variable]] = Field(
        [],
        description="A list of variables to be used when creating templates",
    )
    hub_session_info: Optional[dict] = None
    identity_center_details: Optional[Union[IdentityCenterDetails, None]] = None
    spoke_role_arn: str = Field(
        None,
        description="(Auto-populated) The role arn to assume into when making calls to the account",
    )
    assume_role_arn: Optional[str] = Field(
        None,
        description="The role arn to assume into when making calls to the account",
        exclude=True,
    )

    class Config:
        fields = {"hub_session_info": {"exclude": True}}
        extra = Extra.forbid

    async def get_boto3_session(self, region_name: str = None):
        region_name = region_name or self.region_name

        if self.boto3_session_map is None:
            self.boto3_session_map = {}
        elif boto3_session := self.boto3_session_map.get(region_name):
            return boto3_session

        if not self.hub_session_info:
            await self.set_hub_session_info()

        boto3_session = await create_assume_role_session(
            self.hub_session_info["boto3_session"],
            self.spoke_role_arn,
            region_name,
            external_id=self.external_id,
        )
        if boto3_session:
            self.boto3_session_map[region_name] = boto3_session
            return boto3_session

        return await super().get_boto3_session(region_name)

    async def set_hub_session_info(self):
        region_name = self.region_name
        session = boto3.Session(region_name=region_name)
        if self.aws_profile:
            try:
                session = boto3.Session(
                    profile_name=self.aws_profile, region_name=region_name
                )
            except Exception as err:
                log.warning(err)

        self.hub_session_info = dict(boto3_session=session)
        if self.hub_role_arn and self.hub_role_arn != get_current_role_arn(
            session.client("sts")
        ):
            session = await create_assume_role_session(
                session,
                self.hub_role_arn,
                region_name,
                external_id=self.external_id,
                session_name=os.environ.get("IAMBIC_SESSION_NAME", None),
            )
            if session:
                self.hub_session_info = dict(boto3_session=session)
                return session

        return session

    async def set_identity_center_details(
        self, set_identity_center_map: bool = True, batch_size: int = 35
    ) -> None:
        if self.identity_center_details:
            region = self.identity_center_details.region_name
            identity_center_client = await self.get_boto3_client(
                "sso-admin", region_name=region
            )
            identity_store_client = await self.get_boto3_client(
                "identitystore", region_name=region
            )
            try:
                identity_center_instances = await boto_crud_call(
                    identity_center_client.list_instances
                )
            except errors.AccessDeniedException as err:
                raise Exception(
                    "Please ensure you've specified the correct AWS Identity Center region in "
                    "IAMbic's configuration and that the spoke role has the correct permissions. ",
                    f"Original Exception: {err}",
                )

            if not identity_center_instances.get("Instances"):
                raise ValueError("No Identity Center instances found")

            self.identity_center_details.instance_arn = identity_center_instances[
                "Instances"
            ][0]["InstanceArn"]
            self.identity_center_details.identity_store_id = identity_center_instances[
                "Instances"
            ][0]["IdentityStoreId"]

            if not set_identity_center_map:
                return

            permission_set_arns = await legacy_paginated_search(
                identity_center_client.list_permission_sets,
                response_key="PermissionSets",
                InstanceArn=self.identity_center_details.instance_arn,
            )
            if permission_set_arns:
                # WARNING
                # current implementation does not do well if there is permission set
                # destruction interleave between earlier paginated_search and the sub-
                # sequent describe-permission-set
                permission_set_detail_semaphore = NoqSemaphore(
                    boto_crud_call, batch_size
                )
                permission_set_details = await permission_set_detail_semaphore.process(
                    [
                        {
                            "boto_fnc": identity_center_client.describe_permission_set,
                            "InstanceArn": self.identity_center_details.instance_arn,
                            "PermissionSetArn": permission_set_arn,
                        }
                        for permission_set_arn in permission_set_arns
                    ]
                )
                self.identity_center_details.permission_set_map = {
                    permission_set["PermissionSet"]["Name"]: permission_set[
                        "PermissionSet"
                    ]
                    for permission_set in permission_set_details
                }

            users_and_groups = await asyncio.gather(
                *[
                    legacy_paginated_search(
                        identity_store_client.list_users,
                        response_key="Users",
                        retain_key=True,
                        IdentityStoreId=self.identity_center_details.identity_store_id,
                    ),
                    legacy_paginated_search(
                        identity_store_client.list_groups,
                        response_key="Groups",
                        retain_key=True,
                        IdentityStoreId=self.identity_center_details.identity_store_id,
                    ),
                ]
            )
            for user_or_group in users_and_groups:
                if "Users" in user_or_group:
                    self.identity_center_details.user_map = {
                        user["UserId"]: user for user in user_or_group["Users"]
                    }
                else:
                    self.identity_center_details.group_map = {
                        group["GroupId"]: group for group in user_or_group["Groups"]
                    }

    def dict(
        self,
        *,
        include: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        exclude: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> "DictStrAny":  # noqa
        required_exclude = {
            "boto3_session_map",
            "hub_session_info",
            "identity_center_details",
        }
        if exclude:
            exclude.update(required_exclude)
        else:
            exclude = required_exclude

        resp = super(AWSAccount, self).dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return sort_dict(
            resp,
            [
                "account_id",
                "account_name",
                "org_id",
                "identity_center",
                "default_rule",
            ],
        )

    @property
    def parent_id(self) -> Optional[str]:
        return self.org_id

    @property
    def preferred_identifier(self) -> str:
        return self.account_name

    @property
    def all_identifiers(self) -> set[str]:
        if self.account_name:
            return {self.account_id, self.account_name.lower()}
        else:
            return set(self.account_id)

    def __str__(self):
        return f"{self.account_name} - ({self.account_id})"


class BaseAWSOrgRule(BaseModel):
    iambic_managed: IambicManaged = Field(
        IambicManaged.UNDEFINED,
        description="Controls the directionality of iambic changes",
    )


class AWSOrgAccountRule(AccessModelMixin, BaseAWSOrgRule):
    included_accounts: list[str] = Field(
        ["*"],
        description=(
            "A list of account ids and/or account names this rule applies to. "
            "Account ids/names can be represented as a regex and string"
        ),
    )
    excluded_accounts: Optional[list[str]] = Field(
        [],
        description=(
            "A list of account ids and/or account names this rule explicitly does not apply to. "
            "Account ids/names can be represented as a regex and string"
        ),
    )

    @property
    def included_children(self):
        return self.included_accounts

    @property
    def excluded_children(self):
        return self.excluded_accounts

    @property
    def included_parents(self):
        return []

    @property
    def excluded_parents(self):
        return []


class AWSIdentityCenter(PydanticBaseModel):
    region: Optional[RegionName] = Field(
        RegionName.us_east_1,
        description="Region identity center is configured on",
    )

    @property
    def region_name(self):
        return self.region if isinstance(self.region, str) else self.region.value


class AWSOrganization(BaseAWSAccountAndOrgModel):
    org_name: Optional[str] = Field(
        None,
        description="Optional friendly name for the organization",
        exclude=True,
    )
    org_id: str = Field(
        None,
        description="A unique identifier designating the identity of the organization",
    )
    org_account_id: constr(min_length=12, max_length=12) = Field(
        description="The AWS Organization's master account ID"
    )
    identity_center: Optional[AWSIdentityCenter] = Field(
        default=None,
        description="The AWS Account ID and region of the AWS Identity Center instance to use for this organization",
    )
    default_rule: BaseAWSOrgRule = Field(
        BaseAWSOrgRule(iambic_managed=IambicManaged.UNDEFINED),
        description="The rule used to determine how an organization account should be handled if the account was not found in account_rules.",
    )
    account_rules: Optional[List[AWSOrgAccountRule]] = Field(
        [],
        description="A list of rules used to determine how organization accounts are handled",
    )
    hub_role_arn: str = Field(
        description="The role arn to assume into when making calls to the account",
    )
    spoke_role_is_read_only: bool = Field(
        False,
        description="if true, the spoke role will be limited to read-only permissions",
    )

    async def _create_org_account_instance(
        self, account: dict, session: boto3.Session
    ) -> Optional[AWSAccount]:
        """Create an AWSAccount instance from an AWS Organization account and account dict

        Evaluate rules to determine if the account should be added to the config and if it is a read-only account.
        """
        account_id = account["Id"]
        account_name = account["Name"]
        account_rule = self.default_rule

        if self.account_rules and (
            new_rule := get_provider_value(
                self.account_rules, {account_id, account_name}
            )
        ):
            account_rule = new_rule

        region_name = self.region_name
        if self.identity_center and self.org_account_id == account_id:
            identity_center_details = IdentityCenterDetails(
                region=self.identity_center.region
            )
        else:
            identity_center_details = None

        aws_account = AWSAccount(
            account_id=account_id,
            account_name=account_name,
            org_id=self.org_id,
            variables=account["variables"],
            identity_center_details=identity_center_details,
            iambic_managed=account_rule.iambic_managed,
            spoke_role_arn=get_spoke_role_arn(account_id),
            hub_session_info=dict(boto3_session=session),
            default_region=region_name,
            boto3_session_map={},
        )
        try:
            await aws_account.get_boto3_session()
            return aws_account
        except Exception as err:
            log.debug(
                "Failed to assume role",
                assume_role_arn=aws_account.spoke_role_arn,
                error=err,
            )

    async def get_accounts(self) -> list[AWSAccount]:
        """Get all accounts in an AWS Organization

        Also extends variables for accounts that are already in the config.
        Does not overwrite variables.
        """

        session = await self.get_boto3_session()
        client = await self.get_boto3_client("organizations")
        org_accounts = await legacy_paginated_search(
            client.list_accounts,
            "Accounts",
        )

        active_accounts = [
            account for account in org_accounts if account["Status"] == "ACTIVE"
        ]
        org_accounts = await asyncio.gather(
            *[set_org_account_variables(client, account) for account in active_accounts]
        )
        discovered_accounts = [
            org_account for org_account in org_accounts if org_account
        ]

        discovered_accounts = await asyncio.gather(
            *[
                self._create_org_account_instance(account, session)
                for account in discovered_accounts
            ]
        )
        response = [account for account in discovered_accounts if account]

        if self.identity_center:
            for elem, account in enumerate(response):
                if account.account_id == self.org_account_id:
                    response[elem].identity_center_details.org_account_map = {
                        account["Id"]: account["Name"] for account in active_accounts
                    }

        return response

    def dict(
        self,
        *,
        include: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        exclude: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> "DictStrAny":  # noqa
        if exclude:
            exclude.add("boto3_session_map")
        else:
            exclude = {"boto3_session_map"}

        resp = super(AWSOrganization, self).dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return sort_dict(resp, ["org_id"])

    def __str__(self):
        return self.org_id


class AWSTemplate(BaseTemplate, ExpiryModel):
    identifier: str

    async def _apply_to_account(self, aws_account: AWSAccount) -> AccountChangeDetails:
        raise NotImplementedError

    async def apply(self, config: AWSConfig) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.resource_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type, resource_id=self.resource_id
        )
        relevant_accounts = []

        for account in config.accounts:
            if evaluate_on_provider(self, account):
                relevant_accounts.append(account)
                tasks.append(self._apply_to_account(account))

        if not relevant_accounts:
            if ctx.execute:
                if self.deleted:
                    log_str = "Successfully removed resource."
                    self.delete()
                else:
                    log_str = "No changes detected for resource."
                log.info(log_str, **log_params)

            return template_changes

        if ctx.execute:
            log_str = "Applying changes to resource."
        else:
            log_str = "Detecting changes for resource."

        relevant_accounts_str = [str(account) for account in relevant_accounts]
        log.info(log_str, accounts=relevant_accounts_str, **log_params)

        account_changes: list[AccountChangeDetails] = await asyncio.gather(
            *tasks, return_exceptions=True
        )
        proposed_changes: list[AccountChangeDetails] = []
        exceptions_seen = set()

        for account_change in account_changes:
            if isinstance(account_change, AccountChangeDetails):
                proposed_changes.append(account_change)
            else:
                exceptions_seen.add(str(account_change))

        if exceptions_seen:
            exceptions_seen = list(exceptions_seen)
            proposed_change_accounts = set(
                change.account for change in proposed_changes
            )
            for aws_account in relevant_accounts:
                if str(aws_account) in proposed_change_accounts:
                    continue
                proposed_changes.append(
                    AccountChangeDetails(
                        account=str(aws_account),
                        resource_id=self.resource_id,
                        resource_type=self.resource_type,
                        exceptions_seen=exceptions_seen,
                    )
                )

        template_changes.extend_changes(proposed_changes)

        if template_changes.exceptions_seen:
            if self.deleted:
                cmd_verb = "removing"
            elif ctx.execute:
                cmd_verb = "applying"
            else:
                cmd_verb = "detecting"
            log_str = f"Error encountered when {cmd_verb} resource changes."
        elif account_changes and ctx.execute:
            if self.deleted:
                self.delete()
                log_str = "Successfully removed resource."
            else:
                log_str = "Successfully applied resource changes."
        elif account_changes:
            log_str = "Successfully detected required resource changes."
        else:
            log_str = "No changes detected for resource."

        log.info(log_str, accounts=relevant_accounts_str, **log_params)
        return template_changes

    @property
    def resource_id(self):
        return self.properties.resource_id

    @property
    def resource_type(self):
        return self.properties.resource_type


class Description(AccessModel):
    description: str = ""

    @property
    def resource_type(self) -> str:
        return "aws::description"

    @property
    def resource_id(self) -> str:
        return self.description

    @classmethod
    def new_instance_from_string(cls, s: str) -> Description:
        return Description(description=s)



from __future__ import annotations

import asyncio
import os
import sys
import tempfile
import time
from enum import Enum

from iambic.config.dynamic_config import load_config
from iambic.config.utils import resolve_config_template_path
from iambic.core.models import BaseModel
from iambic.main import run_clone_repos, run_detect, run_git_apply, run_git_plan

REPO_BASE_PATH = os.path.expanduser("~/.iambic/repos/")
PLAN_OUTPUT_PATH = os.environ.get("PLAN_OUTPUT_PATH", None)
FROM_SHA = os.environ.get("FROM_SHA", None)
TO_SHA = os.environ.get("TO_SHA", None)


def init_repo_base_path():
    # Have to be careful when setting REPO_BASE_PATH because lambda execution
    # environment can be very straight on only /tmp is writable

    if os.environ.get("AWS_LAMBDA_FUNCTION_NAME", False):
        temp_templates_directory = tempfile.mkdtemp(prefix="lambda")
        REPO_BASE_PATH = f"{temp_templates_directory}/.iambic/repos/"
    else:
        REPO_BASE_PATH = os.path.expanduser("~/.iambic/repos/")

    this_module = sys.modules[__name__]
    setattr(this_module, "REPO_BASE_PATH", REPO_BASE_PATH)
    os.makedirs(os.path.dirname(REPO_BASE_PATH), exist_ok=True)


def init_plan_output_path():
    # Have to be careful when setting REPO_BASE_PATH because lambda execution
    # environment can be very straight on only /tmp is writable

    if os.environ.get("AWS_LAMBDA_FUNCTION_NAME", False):
        temp_templates_directory = tempfile.mkdtemp(prefix="lambda")
        PLAN_OUTPUT_PATH = f"{temp_templates_directory}/proposed_changes.yaml"
    else:
        PLAN_OUTPUT_PATH = os.environ.get("PLAN_OUTPUT_PATH", None)

    this_module = sys.modules[__name__]
    setattr(this_module, "PLAN_OUTPUT_PATH", PLAN_OUTPUT_PATH)


class LambdaCommand(Enum):
    # `import` is reserved in Python. So we just prepend `run_` to everything.
    run_import = "import"
    run_plan = "plan"
    run_apply = "apply"
    run_detect = "detect"
    run_git_apply = "git_apply"
    run_git_plan = "git_plan"
    run_clone_git_repos = "clone_git_repos"


class LambdaContext(BaseModel):
    command: str


def handler(event, context):
    return run_handler(event, context)


def run_handler(event=None, context=None):
    """
    Default handler for AWS Lambda.
    It is split out from the actual handler, so we can also run via IDE run configurations
    """
    if not context:
        context = {"command": "import"}
    lambda_context = LambdaContext(**context)

    if lambda_context.command == LambdaCommand.run_import.value:
        config_path = asyncio.run(resolve_config_template_path(REPO_BASE_PATH))
        config = asyncio.run(load_config(config_path))
        return asyncio.run(config.run_import(REPO_BASE_PATH))
    elif lambda_context.command == LambdaCommand.run_detect.value:
        return run_detect(REPO_BASE_PATH)
    elif lambda_context.command == LambdaCommand.run_apply.value:
        return run_git_apply(
            False,
            FROM_SHA,
            TO_SHA,
            repo_dir=REPO_BASE_PATH,
            output_path=PLAN_OUTPUT_PATH,
        )
    elif lambda_context.command == LambdaCommand.run_plan.value:
        return run_git_plan(PLAN_OUTPUT_PATH, repo_dir=REPO_BASE_PATH)
    elif lambda_context.command == LambdaCommand.run_git_apply.value:
        return run_git_apply(
            False,
            FROM_SHA,
            TO_SHA,
            repo_dir=REPO_BASE_PATH,
            output_path=PLAN_OUTPUT_PATH,
        )
    elif lambda_context.command == LambdaCommand.run_git_plan.value:
        return run_git_plan(PLAN_OUTPUT_PATH, repo_dir=REPO_BASE_PATH)
    elif lambda_context.command == LambdaCommand.run_clone_git_repos.value:
        return run_clone_repos(REPO_BASE_PATH)
    else:
        raise NotImplementedError(f"Unknown command {lambda_context.command}")


if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise Exception("You must pass a command")
    command = sys.argv[1]

    init_repo_base_path()
    init_plan_output_path()

    # `schedule` is just for development and demo purposes
    if command == "schedule":
        while True:
            run_handler(None, {"command": "clone_git_repos"})
            run_handler(None, {"command": "apply"})
            time.sleep(1)
    else:
        run_handler(None, {"command": command})



from __future__ import annotations

from iambic.core.models import BaseTemplate


class DynamicTemplateRef:
    templates = []

    @classmethod
    def set_templates(cls, templates):
        cls.templates = templates

    @property
    def template_map(self) -> dict[str, BaseTemplate]:
        return {
            template.__fields__["template_type"].default: template
            for template in self.templates
        }


TEMPLATES = DynamicTemplateRef()



from __future__ import annotations

import asyncio
import os
import typing
from typing import TYPE_CHECKING, Any, Optional, Union

import aiohttp
import msal
from pydantic import BaseModel, Field, SecretStr

from iambic.core.context import ctx
from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log
from iambic.core.models import BaseTemplate, TemplateChangeDetails

if TYPE_CHECKING:  # pragma: no cover
    from iambic.plugins.v0_1_0.azure_ad.iambic_plugin import AzureADConfig

    MappingIntStrAny = typing.Mapping[int | str, Any]
    AbstractSetIntStr = typing.AbstractSet[int | str]


class AzureADOrganization(BaseModel):
    idp_name: str
    tenant_id: str
    client_id: str
    client_secret: SecretStr
    request_timeout: int = 60
    client: Any = None
    access_token: str = ""
    iambic_managed: IambicManaged = IambicManaged.UNDEFINED
    require_user_mfa_on_create: bool = Field(
        False,
        description="If true, at next sign-in, the user must perform a multi-factor authentication (MFA) "
        "before being forced to change their password.",
    )

    class Config:
        arbitrary_types_allowed = True

    async def set_azure_access_token(self):  # pragma: no cover
        if not self.access_token:
            # initialize the client here
            self.client = msal.ConfidentialClientApplication(
                self.client_id,
                authority=f"https://login.microsoftonline.com/{self.tenant_id}",
                client_credential=self.client_secret.get_secret_value(),
            )
            token_result = self.client.acquire_token_for_client(
                ["https://graph.microsoft.com/.default"]
            )
            if "access_token" in token_result:
                self.access_token = token_result["access_token"]
            else:
                raise Exception("Access token was not successfully acquired")
        return self.access_token

    async def _make_request(
        self, request_type: str, endpoint: str, **kwargs
    ) -> Union[dict, list, None]:  # pragma: no cover
        await self.set_azure_access_token()

        response = []
        is_list = bool(request_type == "list")
        if is_list:
            request_type = "get"

        async with aiohttp.ClientSession() as session:
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
            }
            url = f"https://graph.microsoft.com/v1.0/{endpoint}"
            while url:
                async with getattr(session, request_type)(
                    url, headers=headers, **kwargs
                ) as resp:
                    if resp.status == 429:
                        # Handle rate limit exceeded error
                        retry_after = int(resp.headers.get("Retry-After", "1"))
                        await asyncio.sleep(retry_after)
                        continue

                    if not resp.ok:
                        log.error(
                            "Azure AD request failed",
                            url=url,
                            org=self.idp_name,
                            message=await resp.text(),
                            status_code=resp.status,
                        )
                        resp.raise_for_status()

                    try:
                        data = await resp.json()
                    except aiohttp.ContentTypeError:
                        return

                    if is_list:
                        response.extend(data["value"])
                        if "@odata.nextLink" in data:
                            url = data["@odata.nextLink"]
                            kwargs.pop(
                                "params", None
                            )  # Clear params since they only apply to the first page
                        else:
                            return response
                    else:
                        return data

    async def post(self, endpoint, **kwargs):
        return await self._make_request("post", endpoint, **kwargs)

    async def get(self, endpoint, **kwargs):
        return await self._make_request("get", endpoint, **kwargs)

    async def list(self, endpoint, **kwargs):
        return await self._make_request("list", endpoint, **kwargs)

    async def patch(self, endpoint, **kwargs):
        return await self._make_request("patch", endpoint, **kwargs)

    async def delete(self, endpoint, **kwargs):
        return await self._make_request("delete", endpoint, **kwargs)

    def dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":  # noqa
        required_exclude = {"access_token", "client"}
        if not exclude:
            exclude = required_exclude
        elif isinstance(exclude, set):
            exclude.update(required_exclude)

        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )


class AzureADTemplate(BaseTemplate):
    idp_name: str = Field(
        ...,
        description="Name of the identity provider that's associated with the resource",
    )

    @property
    def resource_id(self) -> str:
        return self.properties.resource_id

    async def apply(self, config: AzureADConfig) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_id=self.resource_id,
        )

        if self.iambic_managed == IambicManaged.IMPORT_ONLY:
            log_str = "Resource is marked as import only."
            log.info(log_str, **log_params)
            template_changes.proposed_changes = []
            return template_changes

        for azure_ad_organization in config.organizations:
            if azure_ad_organization.idp_name != self.idp_name:
                continue

            if ctx.execute:
                log_str = "Applying changes to resource."
            else:
                log_str = "Detecting changes for resource."
            log.info(log_str, idp_name=azure_ad_organization.idp_name, **log_params)
            tasks.append(self._apply_to_account(azure_ad_organization))

        account_changes = list(await asyncio.gather(*tasks))
        template_changes.extend_changes(account_changes)

        if template_changes.exceptions_seen:
            cmd_verb = "applying" if ctx.execute else "detecting"
            log.error(
                f"Error encountered when {cmd_verb} resource changes.",
                **log_params,
            )
        elif account_changes and ctx.execute:
            if self.deleted:
                self.delete()
                log.info(
                    "Successfully removed resource from all Azure AD organizations.",
                    **log_params,
                )
            else:
                log.info(
                    "Successfully applied resource changes to all Azure AD organizations.",
                    **log_params,
                )
        elif account_changes:
            log.info(
                "Successfully detected required resource changes on all Azure AD organizations.",
                **log_params,
            )
        else:
            log.debug(
                "No changes detected for resource on any Azure AD organization.",
                **log_params,
            )

        return template_changes

    def set_default_file_path(self, repo_dir: str, file_name: str):
        if not file_name.endswith(".yaml"):
            file_name = f"{file_name}.yaml"

        self.file_path = os.path.expanduser(
            os.path.join(
                repo_dir,
                f"resources/{self.resource_type.replace(':', '/')}/{self.idp_name}/{file_name}",
            )
        )



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from googleapiclient import _auth
from googleapiclient.errors import HttpError

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import aio_wrapper

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.google_workspace.group.models import (
        GoogleWorkspaceGroupTemplate,
        GroupMember,
    )
    from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import GoogleProject


async def list_groups(
    domain: str, google_project: GoogleProject
) -> list[GoogleWorkspaceGroupTemplate]:
    from iambic.plugins.v0_1_0.google_workspace.group.models import get_group_template

    groups = []
    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return []
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        raise

    req = await aio_wrapper(service.groups().list, domain=domain)
    res = req.execute(http=http)
    if res and "groups" in res:
        for group in res["groups"]:
            group_template = await get_group_template(service, group, domain)
            groups.append(group_template)
    return groups


async def get_group(group_email: str, domain: str, google_project: GoogleProject):
    from iambic.plugins.v0_1_0.google_workspace.group.models import get_group_template

    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return []
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return

    # TODO: Error handling:
    req = service.groups().get(groupKey=group_email)
    try:
        if group := req.execute(http=http):
            return await get_group_template(service, group, domain)
    except HttpError as err:
        if err.reason == "Not Authorized to access this resource/api":
            log.error("Unable to get group. It may not exist", error=err.reason)
        else:
            raise


async def create_group(
    id: str,
    domain: str,
    email: str,
    name: str,
    description: str,
    google_project: GoogleProject,
):
    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return []
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return
    req = await aio_wrapper(
        service.groups().insert,
        body={"id": id, "email": email, "name": name, "description": description},
    )
    return req.execute(http=http)


async def update_group_domain(
    current_domain: str, proposed_domain: str, log_params: dict[str, str]
):
    response = []
    if current_domain != proposed_domain:
        log_str = "Modifying group domain"
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.UPDATE,
                attribute="domain",
                change_summary={
                    "current_domain": current_domain,
                    "proposed_domain": proposed_domain,
                },
                current_value=current_domain,
                new_value=proposed_domain,
            )
        )
        log.info(
            log_str,
            current_domain=current_domain,
            proposed_domain=proposed_domain,
            **log_params,
        )
        raise NotImplementedError(
            f"Current Domain {current_domain} does not match "
            f"proposed domain {proposed_domain}. We are unable "
            "to update group domains at this point in time."
        )
    return response


async def update_group_description(
    group_email,
    current_description,
    proposed_description,
    domain: str,
    google_project: GoogleProject,
    log_params: dict[str, str],
) -> list[ProposedChange]:
    response: list[ProposedChange] = []
    if current_description == proposed_description:
        return response
    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return response
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return response
    log_str = "Detected updated group description"
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=group_email,
            resource_type="google:group:template",
            attribute="description",
            change_summary={
                "current_description": current_description,
                "proposed_description": proposed_description,
            },
            current_value=current_description,
            new_value=proposed_description,
        )
    )
    if ctx.execute:
        log_str = "Updating group description"
        await aio_wrapper(
            service.groups()
            .patch(groupKey=group_email, body={"description": proposed_description})
            .execute,
            http=http,
        )
    log.info(
        log_str,
        current_description=current_description,
        proposed_description=proposed_description,
        **log_params,
    )
    return response


async def update_group_name(
    group_email: str,
    current_name: str,
    proposed_name: str,
    domain: str,
    google_project: GoogleProject,
    log_params: dict[str, str],
) -> list[ProposedChange]:
    response: list[ProposedChange] = []
    if current_name == proposed_name:
        return response
    log_str = "Detected group name update"
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=group_email,
            resource_type="google:group:template",
            attribute="group_name",
            current_value=current_name,
            new_value=proposed_name,
        )
    )

    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return response
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return response
    if ctx.execute:
        log_str = "Updating group name"
        await aio_wrapper(
            service.groups()
            .patch(groupKey=group_email, body={"name": proposed_name})
            .execute,
            http=http,
        )
    log.info(
        log_str, current_name=current_name, proposed_name=proposed_name, **log_params
    )
    return response


async def update_group_email(
    current_email: str,
    proposed_email: str,
    domain: str,
    google_project: GoogleProject,
    log_params: dict[str, str],
) -> list[ProposedChange]:
    # TODO: This won't work as-is, since we aren't really aware of the old e-mail
    response: list[ProposedChange] = []
    if current_email == proposed_email:
        return response
    log_str = "Detected group e-mail update"
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=current_email,
            resource_type="google:group:template",
            attribute="group_email",
            current_value=current_email,
            new_value=proposed_email,
        )
    )

    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return response
        http = _auth.authorized_http(service._http.credentials)
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return response
    if ctx.execute:
        log_str = "Updating group e-mail"
        await aio_wrapper(
            service.groups()
            .patch(groupKey=current_email, body={"email": proposed_email})
            .execute,
            http=http,
        )
    log.info(
        log_str, current_name=current_email, proposed_name=proposed_email, **log_params
    )
    return response


async def maybe_delete_group(
    group: GoogleWorkspaceGroupTemplate,
    google_project: GoogleProject,
    log_params: dict[str, str],
) -> list[ProposedChange]:
    response: list[ProposedChange] = []
    if not group.deleted:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=group.properties.email,
            resource_type=group.resource_type,
            attribute="group",
            change_summary={"group": group.properties.name},
            current_value=group.properties.name,
        )
    )
    if ctx.execute:
        try:
            service = await google_project.get_service_connection(
                "admin", "directory_v1", group.properties.domain
            )
            if not service:
                return []
            http = _auth.authorized_http(service._http.credentials)
        except AttributeError as err:
            log.exception("Unable to process google groups.", error=err)
            return []

        log_str = "Detected group deletion"

        if ctx.execute:
            log_str = "Deleting Group"
            await aio_wrapper(
                service.groups().delete(groupKey=group.properties.email).execute,
                http=http,
            )
        log.info(log_str, group_email=group.properties.email, **log_params)
    return response


async def get_group_members(service, group):
    from iambic.plugins.v0_1_0.google_workspace.group.models import (
        GroupMember,
        GroupMemberRole,
        GroupMemberStatus,
        GroupMemberType,
    )

    http = _auth.authorized_http(service._http.credentials)
    member_req = service.members().list(groupKey=group["email"])
    member_res = member_req.execute(http=http) or {}
    return [
        GroupMember(
            email=member["email"],
            role=GroupMemberRole(member["role"]),
            type=GroupMemberType(member["type"]),
            status=GroupMemberStatus(member.get("status", GroupMemberStatus.UNDEFINED)),
        )
        for member in member_res.get("members", [])
    ]


async def update_group_members(
    group_email: str,
    current_members: list[GroupMember],
    proposed_members: list[GroupMember],
    domain: str,
    google_project: GoogleProject,
    log_params: dict[str, str],
):
    # TODO: This will likely fail if I change the Role of a user, since we are doing all
    # of these operations with asyncio.gather. Should do the remove operations first, then the add ones.
    tasks = []
    response = []
    try:
        service = await google_project.get_service_connection(
            "admin", "directory_v1", domain
        )
        if not service:
            return False
    except AttributeError as err:
        log.exception("Unable to process google groups.", error=err)
        return

    if users_to_remove := [
        member
        for member in current_members
        if member.email not in [m.email for m in proposed_members]
    ]:
        log_str = "Detected users to remove from group"
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_id=group_email,
                resource_type="google:group:template",
                attribute="users",
                change_summary={
                    "UsersToRemove": [user.email for user in users_to_remove]
                },
                current_value=[user.email for user in current_members],
                new_value=[user.email for user in proposed_members],
            )
        )
        if ctx.execute:
            log_str = "Removing users from group"
            for user in users_to_remove:
                http = _auth.authorized_http(service._http.credentials)
                tasks.append(
                    aio_wrapper(
                        service.members()
                        .delete(groupKey=group_email, memberKey=user.email)
                        .execute,
                        http=http,
                    )
                )
        log.info(log_str, users=[user.email for user in users_to_remove], **log_params)

    if users_to_add := [
        member
        for member in proposed_members
        if member.email not in [m.email for m in current_members]
    ]:
        log_str = "Detected new users to add to group"
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_id=group_email,
                resource_type="google:group:template",
                attribute="users",
                change_summary={"UsersToAdd": [user.email for user in users_to_add]},
                current_value=[user.email for user in current_members],
                new_value=[user.email for user in proposed_members],
            )
        )
        if ctx.execute:
            log_str = "Adding users to group"
            for user in users_to_add:
                http = _auth.authorized_http(service._http.credentials)
                tasks.append(
                    aio_wrapper(
                        service.members()
                        .insert(
                            groupKey=group_email,
                            body={
                                "email": user.email,
                                "role": user.role.value,
                                "type": user.type.value,
                                "status": user.status.value,
                            },
                        )
                        .execute,
                        http=http,
                    )
                )

        log.info(log_str, users=[user.email for user in users_to_add], **log_params)
    if tasks:
        await asyncio.gather(*tasks)
    return response



from __future__ import annotations

import asyncio

from iambic.core.logger import log
from iambic.core.parser import load_templates
from iambic.core.utils import remove_expired_resources


async def flag_expired_resources(template_paths: list[str]):
    # Warning: The dynamic config must be loaded before this is called.
    #   This is done using iambic.config.dynamic_config.load_config(config_path)
    log.info("Scanning for expired resources")
    templates = await asyncio.gather(
        *[
            remove_expired_resources(
                template, template.resource_type, template.resource_id
            )
            for template in load_templates(template_paths)
        ]
    )

    for template in templates:
        template.write(exclude_none=True, exclude_unset=True, exclude_defaults=True)

    log.info("Expired resource scan complete.")



from typing import Union

from iambic.core.template_generation import (
    base_group_str_attribute as core_base_group_str_attribute,
)
from iambic.core.template_generation import (
    group_dict_attribute as core_group_dict_attribute,
)
from iambic.core.template_generation import (
    group_int_or_str_attribute as core_group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount


async def base_group_str_attribute(
    aws_account_map: dict[str, AWSAccount], account_resources: list[dict]
) -> dict[str, list]:
    """Groups a string attribute by a shared name across aws_accounts

    Call group_int_or_str_attribute instead unless you need to transform this response.
    An example would be grouping role names for generating the template where you need to keep the file_path ref.
    :param aws_account_map: dict(account_id:str = AWSAccount)
    :param account_resources: list[dict(account_id:str, resources=list[dict(resource_val: str, **)])]
    :return: dict(attribute_val: str = list[dict(resource_val: str, account_id: str, **)])
    """

    return await core_base_group_str_attribute(
        aws_account_map, account_resources, "account_id"
    )


async def group_int_or_str_attribute(
    aws_account_map: dict[str, AWSAccount],
    number_of_accounts_resource_on: int,
    account_resources: Union[dict, list[dict]],
    key: Union[int, str],
) -> Union[int, str, list[dict]]:
    """Groups an attribute by aws_accounts, formats the attribute and normalizes the included aws_accounts.
    :param aws_account_map:
    :param number_of_accounts_resource_on:
    :param account_resources: dict(account_id: str = int_val) , list[dict(account_id:str, resources=list[dict])]
    :param key: Used to form the list[dict] response when there are multiple values for the attribute.
    :return:
    """

    return await core_group_int_or_str_attribute(
        aws_account_map,
        number_of_accounts_resource_on,
        account_resources,
        "account_id",
        "included_accounts",
        key,
    )


async def group_dict_attribute(
    aws_account_map: dict[str, AWSAccount],
    number_of_accounts_resource_on: int,
    account_resources: list[dict],
    is_dict_attr: bool = True,
    prefer_templatized: bool = False,
) -> Union[dict, list[dict]]:
    """Groups an attribute by aws_accounts, formats the attribute and normalizes the included aws_accounts.
    :param aws_account_map: {account_id: aws_account}
    :param number_of_accounts_resource_on:
    :param account_resources: list[dict(account_id:str, resources=list[dict])]
    :param is_dict_attr: If false and only one hit, still return as a list. Useful for things like inline_policies.
    :return:
    """

    return await core_group_dict_attribute(
        aws_account_map,
        number_of_accounts_resource_on,
        account_resources,
        "account_id",
        "included_accounts",
        is_dict_attr,
        prefer_templatized,
    )



from __future__ import annotations

import asyncio
import os
from datetime import datetime, timedelta
from typing import Union

from iambic.core.logger import log
from iambic.plugins.v0_1_0.aws.models import (
    IAMBIC_HUB_ROLE_NAME,
    IAMBIC_SPOKE_ROLE_NAME,
    get_hub_role_arn,
)
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, legacy_paginated_search

TEMPLATE_DIR = f"{str(os.path.dirname(__file__))}/templates"


def get_rule_forwarding_template_body() -> str:
    template = f"{TEMPLATE_DIR}/IdentityRuleForwarder.yml"
    with open(template, "r") as f:
        return f.read()


def get_central_rule_template_body() -> str:
    template = f"{TEMPLATE_DIR}/IdentityRuleDestination.yml"
    with open(template, "r") as f:
        return f.read()


def get_iambic_hub_role_template_body() -> str:
    template = f"{TEMPLATE_DIR}/IambicHubRole.yml"
    with open(template, "r") as f:
        return f.read()


def get_iambic_spoke_role_template_body(read_only=False) -> str:
    suffix = "ReadOnly" if read_only else ""
    template = f"{TEMPLATE_DIR}/IambicSpokeRole{suffix}.yml"
    with open(template, "r") as f:
        return f.read()


async def create_stack(
    client, stack_name: str, template_body: str, parameters: list[dict], **kwargs
) -> bool:
    region = client.meta.region_name

    try:
        existing_stack = await legacy_paginated_search(
            client.describe_stacks,
            response_key="Stacks",
            StackName=stack_name,
        )
        if existing_stack:
            log.info(
                "Stack already exists. Skipping creation. "
                "If this is not the desired behavior, please delete the stack in the console and try again.\n",
                stack_name=stack_name,
                stack_url=f"https://{region}.console.aws.amazon.com/cloudformation/home?region={region}#/stacks\n",
            )
            return True
    except client.exceptions.ClientError as e:
        # ValidationError means the stack doesn't exist
        if e.response["Error"]["Code"] != "ValidationError":
            raise

    response = await boto_crud_call(
        client.create_stack,
        StackName=stack_name,
        TemplateBody=template_body,
        Parameters=parameters,
        OnFailure="ROLLBACK",
        **kwargs,
    )

    stack_id = response.get("StackId")
    resource_status = "CREATE_IN_PROGRESS"

    estimated_completion = (datetime.now() + timedelta(minutes=3)).strftime(
        "%m/%d/%Y, %H:%M:%S"
    )

    while resource_status == "CREATE_IN_PROGRESS":
        log.info(
            "Waiting for stack to be created.",
            stack_name=stack_name,
            estimated_completion=estimated_completion,
        )
        await asyncio.sleep(30)
        response = await legacy_paginated_search(
            client.describe_stacks,
            response_key="Stacks",
            StackName=stack_name,
        )
        stack_details = [stack for stack in response if stack["StackId"] == stack_id]
        if stack_details:
            stack_details = stack_details[0]
        else:
            stack_details = {
                "StackStatus": "CREATE_FAILED",
                "StackStatusReason": f"Stack not found. Stack Name: {stack_name}, Stack ID: {stack_id}",
            }
        resource_status = stack_details.get("StackStatus")

    if resource_status == "CREATE_FAILED":
        soft_fail = bool("already exists." in stack_details.get("StackStatusReason"))
        if soft_fail:
            log.warning(
                "Role already exists on account.",
                stack_name=stack_name,
                reason=stack_details.get("StackStatusReason"),
            )
        else:
            log.error(
                "Unable to create AWS CloudFormation Stack",
                stack_name=stack_name,
                reason=stack_details.get("StackStatusReason"),
            )
        return soft_fail
    return True


async def create_stack_set(
    client,
    stack_set_name,
    template_body: str,
    parameters: list[dict],
    deployment_targets: dict,
    deployment_regions: list[str],
    operation_preferences: dict[str, Union[int, str]],
    **kwargs,
) -> bool:
    region = client.meta.region_name

    try:
        await boto_crud_call(
            client.describe_stack_set,
            StackSetName=stack_set_name,
        )
        log.info(
            "StackSet already exists. Skipping creation. "
            "If this is not the desired behavior, please delete the stack in the console and try again.\n",
            stack_set_name=stack_set_name,
            stack_set_url=f"https://{region}.console.aws.amazon.com/cloudformation/home?region={region}#/stacksets\n",
        )
        return True
    except client.exceptions.StackSetNotFoundException:
        pass

    await boto_crud_call(
        client.create_stack_set,
        StackSetName=stack_set_name,
        TemplateBody=template_body,
        Parameters=parameters,
        PermissionModel="SERVICE_MANAGED",
        AutoDeployment={"Enabled": True, "RetainStacksOnAccountRemoval": True},
        ManagedExecution={"Active": True},
        **kwargs,
    )

    await boto_crud_call(
        client.create_stack_instances,
        StackSetName=stack_set_name,
        DeploymentTargets=deployment_targets,
        Regions=deployment_regions,
        OperationPreferences=operation_preferences,
    )

    await asyncio.sleep(30)  # Wait for stack instances to be created

    while True:
        stack_instances = await legacy_paginated_search(
            client.list_stack_instances,
            response_key="Summaries",
            StackSetName=stack_set_name,
        )
        pending_count = len(
            [
                instance
                for instance in stack_instances
                if instance.get("StackInstanceStatus", {}).get("DetailedStatus")
                in ["PENDING", "RUNNING"]
            ]
        )
        if pending_count:
            log.info(
                "Waiting for stack set instances to be created",
                stack_set_name=stack_set_name,
                pending_instances=pending_count,
                estimated_completion=(
                    datetime.now() + timedelta(minutes=pending_count * 1.5)
                ).strftime("%m/%d/%Y, %H:%M:%S"),
            )
            await asyncio.sleep(30)
            continue
        elif all(
            instance.get("StackInstanceStatus", {}).get("DetailedStatus") == "SUCCEEDED"
            for instance in stack_instances
        ):
            return True
        else:
            existing_instances = []
            failed_instances = []
            for instance in stack_instances:
                if instance.get("StackInstanceStatus", {}).get("DetailedStatus") in [
                    "FAILED",
                    "CANCELLED",
                ]:
                    response_summary = {
                        "Account": instance["Account"],
                        "Region": instance["Region"],
                        "Status": instance.get("StackInstanceStatus", {}).get(
                            "DetailedStatus"
                        ),
                        "StatusReason": instance.get("StatusReason"),
                    }
                    if "already exists." in instance.get("StatusReason"):
                        existing_instances.append(response_summary)
                    else:
                        failed_instances.append(response_summary)

            if existing_instances:
                log.warning(
                    "Role already exists on account(s).",
                    stack_set_name=stack_set_name,
                    failed_instances=failed_instances,
                )
            if failed_instances:
                log.error(
                    "Unable to create stack set instances",
                    stack_set_name=stack_set_name,
                    failed_instances=failed_instances,
                )

            return not bool(failed_instances)


async def create_change_detection_stacks(
    cf_client, org_id: str, org_account_id: str, role_arn: str = None
) -> bool:
    region = cf_client.meta.region_name
    additional_kwargs = {"RoleARN": role_arn} if role_arn else {}
    stack_created = await create_stack(
        cf_client,
        stack_name="IAMbicCentralChangeRule",
        template_body=get_central_rule_template_body(),
        parameters=[{"ParameterKey": "OrgID", "ParameterValue": org_id}],
        **additional_kwargs,
    )
    if stack_created:
        stack_created = await create_stack(
            cf_client,
            stack_name="IAMbicForwardEventRule",
            template_body=get_rule_forwarding_template_body(),
            parameters=[
                {
                    "ParameterKey": "TargetEventBusArn",
                    "ParameterValue": f"arn:aws:events:{region}:{org_account_id}:event-bus/IAMbicChangeDetectionEventBus",
                }
            ],
            Capabilities=["CAPABILITY_NAMED_IAM"],
            **additional_kwargs,
        )

    return stack_created


async def create_change_detection_stack_sets(
    cf_client, org_client, org_account_id: str
) -> bool:
    region = cf_client.meta.region_name
    org_roots = await legacy_paginated_search(
        org_client.list_roots, response_key="Roots"
    )

    return await create_stack_set(
        cf_client,
        stack_set_name="IAMbicForwardEventRule",
        template_body=get_rule_forwarding_template_body(),
        parameters=[
            {
                "ParameterKey": "TargetEventBusArn",
                "ParameterValue": f"arn:aws:events:{region}:{org_account_id}:event-bus/IAMbicChangeDetectionEventBus",
            }
        ],
        deployment_targets={
            "OrganizationalUnitIds": [root["Id"] for root in org_roots],
            "AccountFilterType": "NONE",
        },
        deployment_regions=[region],
        operation_preferences={
            "MaxConcurrentCount": 1,
            "FailureToleranceCount": 1,
        },
        Capabilities=["CAPABILITY_NAMED_IAM"],
    )


async def create_iambic_eventbridge_stacks(
    cf_client, org_client, org_id: str, account_id: str, role_arn: str = None
) -> bool:
    region = cf_client.meta.region_name
    successfully_created = await create_change_detection_stacks(
        cf_client, org_id, account_id, role_arn
    )
    if successfully_created:
        log.info(
            f"WARNING: Do not exit; creating stack instances.\n"
            f"You can check the progress here:\n"
            f"https://{region}.console.aws.amazon.com/cloudformation/home?region={region}#/stacksets/IAMbicForwardEventRule/stacks\n"
        )
        return await create_change_detection_stack_sets(
            cf_client, org_client, account_id
        )

    return successfully_created


async def create_spoke_role_stack_set(
    cf_client,
    org_client,
    hub_account_id: str,
    read_only=False,
) -> bool:
    region = cf_client.meta.region_name
    org_roots = await legacy_paginated_search(
        org_client.list_roots, response_key="Roots"
    )

    return await create_stack_set(
        cf_client,
        stack_set_name="IambicSpokeRole",
        template_body=get_iambic_spoke_role_template_body(read_only=read_only),
        parameters=[
            {
                "ParameterKey": "HubRoleArn",
                "ParameterValue": get_hub_role_arn(hub_account_id),
            },
            {"ParameterKey": "SpokeRoleName", "ParameterValue": IAMBIC_SPOKE_ROLE_NAME},
        ],
        deployment_targets={
            "OrganizationalUnitIds": [root["Id"] for root in org_roots],
            "AccountFilterType": "NONE",
        },
        deployment_regions=[region],
        operation_preferences={
            "RegionConcurrencyType": "PARALLEL",
            "MaxConcurrentCount": 10,
            "FailureToleranceCount": 10,
        },
        Capabilities=["CAPABILITY_NAMED_IAM"],
    )


async def create_spoke_role_stack(
    cf_client,
    hub_account_id: str,
    role_arn: str = None,
    read_only=False,
) -> bool:
    additional_kwargs = {"RoleARN": role_arn} if role_arn else {}
    return await create_stack(
        cf_client,
        stack_name="IambicSpokeRole",
        template_body=get_iambic_spoke_role_template_body(read_only=read_only),
        parameters=[
            {
                "ParameterKey": "HubRoleArn",
                "ParameterValue": get_hub_role_arn(hub_account_id),
            },
            {
                "ParameterKey": "SpokeRoleName",
                "ParameterValue": IAMBIC_SPOKE_ROLE_NAME,
            },
        ],
        Capabilities=["CAPABILITY_NAMED_IAM"],
        **additional_kwargs,
    )


async def create_hub_role_stack(
    cf_client,
    hub_account_id: str,
    assume_as_arn: str,
    role_arn: str = None,
) -> bool:
    additional_kwargs = {"RoleARN": role_arn} if role_arn else {}
    stack_created = await create_stack(
        cf_client,
        stack_name="IambicHubRole",
        template_body=get_iambic_hub_role_template_body(),
        parameters=[
            {"ParameterKey": "HubRoleName", "ParameterValue": IAMBIC_HUB_ROLE_NAME},
            {
                "ParameterKey": "SpokeRoleName",
                "ParameterValue": IAMBIC_SPOKE_ROLE_NAME,
            },
            {"ParameterKey": "AssumeAsArn", "ParameterValue": assume_as_arn},
        ],
        Capabilities=["CAPABILITY_NAMED_IAM"],
        **additional_kwargs,
    )
    if stack_created:
        return await create_spoke_role_stack(cf_client, hub_account_id, role_arn)

    return stack_created


async def create_iambic_role_stacks(
    cf_client,
    hub_account_id: str,
    assume_as_arn: str,
    role_arn: str = None,
    org_client=None,
    read_only=False,
) -> bool:
    hub_role_created = await create_hub_role_stack(
        cf_client,
        hub_account_id,
        assume_as_arn,
        role_arn,
    )
    region = cf_client.meta.region_name
    if hub_role_created and org_client:
        log.info(
            f"WARNING: Do not exit; creating stack instances.\n"
            f"You can check the progress here:\n"
            f"https://{region}.console.aws.amazon.com/cloudformation/home?region={region}#/stacksets/IambicSpokeRole/stacks\n"
        )
        return await create_spoke_role_stack_set(
            cf_client, org_client, hub_account_id, read_only=read_only
        )

    return hub_role_created
